import io
import json
import os
import sys
import threading
import unittest
import unittest.mock as mock

import paramiko
import requests

from gradio import Interface, networking, tunneling

os.environ["GRADIO_ANALYTICS_ENABLED"] = "False"


class TestTunneling(unittest.TestCase):
    def test_create_tunnel(self):
        response = requests.get(networking.GRADIO_API_SERVER)
        payload = response.json()[0]
        io = Interface(lambda x: x, "text", "text")
        _, path_to_local_server, _ = io.launch(prevent_thread_lock=True, share=False)
        _, localhost, port = path_to_local_server.split(":")
        threading.Thread.start = mock.MagicMock(return_value=None)
        paramiko.SSHClient.connect = mock.MagicMock(return_value=None)
        tunneling.create_tunnel(payload, localhost, port)
        threading.Thread.start.assert_called_once()
        paramiko.SSHClient.connect.assert_called_once()
        io.close()


class TestVerbose(unittest.TestCase):
    """Not absolutely needed but just including them for the sake of completion."""

    def setUp(self):
        self.message = "print test"
        self.capturedOutput = io.StringIO()  # Create StringIO object
        sys.stdout = self.capturedOutput  # and redirect stdout.

    def test_verbose_debug_true(self):
        tunneling.verbose(self.message, debug_mode=True)
        self.assertEqual(self.capturedOutput.getvalue().strip(), self.message)

    def test_verbose_debug_false(self):
        tunneling.verbose(self.message, debug_mode=False)
        self.assertEqual(self.capturedOutput.getvalue().strip(), "")

    def tearDown(self):
        sys.stdout = sys.__stdout__


if __name__ == "__main__":
    unittest.main()



"""
This is the core file in the `gradio` package, and defines the Interface class,
including various methods for constructing an interface and then launching it.
"""

from __future__ import annotations

import copy
import csv
import inspect
import json
import os
import random
import re
import warnings
import weakref
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple

from markdown_it import MarkdownIt
from mdit_py_plugins.footnote import footnote_plugin

from gradio import interpretation, utils
from gradio.blocks import Blocks, Column, Row, TabItem, Tabs
from gradio.components import (
    Button,
    Component,
    Dataset,
    Interpretation,
    Markdown,
    StatusTracker,
    Variable,
    get_component_instance,
)
from gradio.external import load_from_pipeline, load_interface  # type: ignore
from gradio.flagging import CSVLogger, FlaggingCallback  # type: ignore
from gradio.process_examples import cache_interface_examples, load_from_cache

if TYPE_CHECKING:  # Only import for type checking (is False at runtime).
    import transformers


class Interface(Blocks):
    """
    Gradio interfaces are created by constructing a `Interface` object
    with a locally-defined function, or with `Interface.load()` with the path
    to a repo or by `Interface.from_pipeline()` with a Transformers Pipeline.
    """

    # stores references to all currently existing Interface instances
    instances: weakref.WeakSet = weakref.WeakSet()

    class InterfaceTypes(Enum):
        STANDARD = auto()
        INPUT_ONLY = auto()
        OUTPUT_ONLY = auto()
        UNIFIED = auto()

    @classmethod
    def get_instances(cls) -> List[Interface]:
        """
        :return: list of all current instances.
        """
        return list(Interface.instances)

    @classmethod
    def load(
        cls,
        name: str,
        src: Optional[str] = None,
        api_key: Optional[str] = None,
        alias: Optional[str] = None,
        **kwargs,
    ) -> Interface:
        """
        Class method to construct an Interface from an external source repository, such as huggingface.
        Parameters:
        name (str): the name of the model (e.g. "gpt2"), can include the `src` as prefix (e.g. "huggingface/gpt2")
        src (str): the source of the model: `huggingface` or `gradio` (or empty if source is provided as a prefix in `name`)
        api_key (str): optional api key for use with Hugging Face Model Hub
        alias (str): optional, used as the name of the loaded model instead of the default name
        Returns:
        (gradio.Interface): a Gradio Interface object for the given model
        """
        interface_info = load_interface(name, src, api_key, alias)
        kwargs = dict(interface_info, **kwargs)
        interface = cls(**kwargs)
        interface.api_mode = True  # So interface doesn't run pre/postprocess.
        return interface

    @classmethod
    def from_pipeline(cls, pipeline: transformers.Pipeline, **kwargs) -> Interface:
        """
        Construct an Interface from a Hugging Face transformers.Pipeline.
        Parameters:
        pipeline (transformers.Pipeline):
        Returns:
        (gradio.Interface): a Gradio Interface object from the given Pipeline

        Example usage:
            import gradio as gr
            from transformers import pipeline
            pipe = pipeline(model="lysandre/tiny-vit-random")
            gr.Interface.from_pipeline(pipe).launch()
        """
        interface_info = load_from_pipeline(pipeline)
        kwargs = dict(interface_info, **kwargs)
        interface = cls(**kwargs)
        return interface

    def __init__(
        self,
        fn: Callable | List[Callable],
        inputs: Optional[str | Component | List[str | Component]] = None,
        outputs: Optional[str | Component | List[str | Component]] = None,
        examples: Optional[List[Any] | List[List[Any]] | str] = None,
        cache_examples: Optional[bool] = None,
        examples_per_page: int = 10,
        live: bool = False,
        layout: str = "unaligned",
        show_input: bool = True,
        show_output: bool = True,
        interpretation: Optional[Callable | str] = None,
        num_shap: float = 2.0,
        theme: Optional[str] = None,
        repeat_outputs_per_model: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        article: Optional[str] = None,
        thumbnail: Optional[str] = None,
        css: Optional[str] = None,
        allow_flagging: Optional[str] = None,
        flagging_options: List[str] = None,
        flagging_dir: str = "flagged",
        analytics_enabled: Optional[bool] = None,
        flagging_callback: FlaggingCallback = CSVLogger(),
        **kwargs,
    ):
        """
        Parameters:
        fn (Union[Callable, List[Callable]]): the function to wrap an interface around.
        inputs (Union[str, InputComponent, List[Union[str, InputComponent]]]): a single Gradio input component, or list of Gradio input components. Components can either be passed as instantiated objects, or referred to by their string shortcuts. The number of input components should match the number of parameters in fn.
        outputs (Union[str, OutputComponent, List[Union[str, OutputComponent]]]): a single Gradio output component, or list of Gradio output components. Components can either be passed as instantiated objects, or referred to by their string shortcuts. The number of output components should match the number of values returned by fn.
        examples (Union[List[List[Any]], str]): sample inputs for the function; if provided, appears below the UI components and can be used to populate the interface. Should be nested list, in which the outer list consists of samples and each inner list consists of an input corresponding to each input component. A string path to a directory of examples can also be provided. If there are multiple input components and a directory is provided, a log.csv file must be present in the directory to link corresponding inputs.
        examples_per_page (int): If examples are provided, how many to display per page.
        cache_examples(Optional[bool]):
            If True, caches examples in the server for fast runtime in examples.
            The default option in HuggingFace Spaces is True.
            The default option elsewhere is False.
        live (bool): whether the interface should automatically reload on change.
        layout (str): Layout of input and output panels. "horizontal" arranges them as two columns of equal height, "unaligned" arranges them as two columns of unequal height, and "vertical" arranges them vertically.
        interpretation (Union[Callable, str]): function that provides interpretation explaining prediction output. Pass "default" to use simple built-in interpreter, "shap" to use a built-in shapley-based interpreter, or your own custom interpretation function.
        num_shap (float): a multiplier that determines how many examples are computed for shap-based interpretation. Increasing this value will increase shap runtime, but improve results. Only applies if interpretation is "shap".
        title (str): a title for the interface; if provided, appears above the input and output components.
        description (str): a description for the interface; if provided, appears above the input and output components.
        article (str): an expanded article explaining the interface; if provided, appears below the input and output components. Accepts Markdown and HTML content.
        thumbnail (str): path to image or src to use as display picture for models listed in gradio.app/hub
        theme (str): Theme to use - one of "default", "huggingface", "seafoam", "grass", "peach". Add "dark-" prefix, e.g. "dark-peach" for dark theme (or just "dark" for the default dark theme).
        css (str): custom css or path to custom css file to use with interface.
        allow_flagging (str): one of "never", "auto", or "manual". If "never" or "auto", users will not see a button to flag an input and output. If "manual", users will see a button to flag. If "auto", every prediction will be automatically flagged. If "manual", samples are flagged when the user clicks flag button. Can be set with environmental variable GRADIO_ALLOW_FLAGGING.
        flagging_options (List[str]): if provided, allows user to select from the list of options when flagging. Only applies if allow_flagging is "manual".
        flagging_dir (str): what to name the dir where flagged data is stored.
        """
        super().__init__(
            analytics_enabled=analytics_enabled, mode="interface", **kwargs
        )

        self.interface_type = self.InterfaceTypes.STANDARD
        if inputs is None and outputs is None:
            raise ValueError("Must provide at least one of `inputs` or `outputs`")
        elif outputs is None:
            outputs = []
            self.interface_type = self.InterfaceTypes.INPUT_ONLY
        elif inputs is None:
            inputs = []
            self.interface_type = self.InterfaceTypes.OUTPUT_ONLY

        if not isinstance(fn, list):
            fn = [fn]
        if not isinstance(inputs, list):
            inputs = [inputs]
        if not isinstance(outputs, list):
            outputs = [outputs]

        if self.is_space and cache_examples is None:
            self.cache_examples = True
        else:
            self.cache_examples = cache_examples or False

        if "state" in inputs or "state" in outputs:
            state_input_count = len([i for i in inputs if i == "state"])
            state_output_count = len([o for o in outputs if o == "state"])
            if state_input_count != 1 or state_output_count != 1:
                raise ValueError(
                    "If using 'state', there must be exactly one state input and one state output."
                )
            default = utils.get_default_args(fn[0])[inputs.index("state")]
            state_variable = Variable(default_value=default)
            inputs[inputs.index("state")] = state_variable
            outputs[outputs.index("state")] = state_variable

            if cache_examples:
                warnings.warn(
                    "Cache examples cannot be used with state inputs and outputs."
                    "Setting cache_examples to False."
                )
            self.cache_examples = False

        self.input_components = [get_component_instance(i).unrender() for i in inputs]
        self.output_components = [get_component_instance(o).unrender() for o in outputs]

        if len(self.input_components) == len(self.output_components):
            same_components = [
                i is o for i, o in zip(self.input_components, self.output_components)
            ]
            if all(same_components):
                self.interface_type = self.InterfaceTypes.UNIFIED

        if self.interface_type in [
            self.InterfaceTypes.STANDARD,
            self.InterfaceTypes.OUTPUT_ONLY,
        ]:
            for o in self.output_components:
                o.interactive = False  # Force output components to be non-interactive

        if repeat_outputs_per_model:
            self.output_components *= len(fn)

        if (
            interpretation is None
            or isinstance(interpretation, list)
            or callable(interpretation)
        ):
            self.interpretation = interpretation
        elif isinstance(interpretation, str):
            self.interpretation = [
                interpretation.lower() for _ in self.input_components
            ]
        else:
            raise ValueError("Invalid value for parameter: interpretation")

        self.api_mode = False
        self.predict = fn
        self.predict_durations = [[0, 0]] * len(fn)
        self.function_names = [func.__name__ for func in fn]
        self.__name__ = ", ".join(self.function_names)

        self.live = live
        self.layout = layout
        self.show_input = show_input
        self.show_output = show_output
        self.flag_hash = random.getrandbits(32)

        self.title = title

        CLEANER = re.compile("<.*?>")

        def clean_html(raw_html):
            cleantext = re.sub(CLEANER, "", raw_html)
            return cleantext

        md = MarkdownIt(
            "js-default",
            {
                "linkify": True,
                "typographer": True,
                "html": True,
            },
        ).use(footnote_plugin)

        simple_description = None
        if description is not None:
            description = md.render(description)
            simple_description = clean_html(description)
        self.simple_description = simple_description
        self.description = description
        if article is not None:
            article = utils.readme_to_html(article)
            article = md.render(article)
        self.article = article

        self.thumbnail = thumbnail
        theme = theme if theme is not None else os.getenv("GRADIO_THEME", "default")
        DEPRECATED_THEME_MAP = {
            "darkdefault": "default",
            "darkhuggingface": "dark-huggingface",
            "darkpeach": "dark-peach",
            "darkgrass": "dark-grass",
        }
        VALID_THEME_SET = (
            "default",
            "huggingface",
            "seafoam",
            "grass",
            "peach",
            "dark",
            "dark-huggingface",
            "dark-seafoam",
            "dark-grass",
            "dark-peach",
        )
        if theme in DEPRECATED_THEME_MAP:
            warnings.warn(
                f"'{theme}' theme name is deprecated, using {DEPRECATED_THEME_MAP[theme]} instead."
            )
            theme = DEPRECATED_THEME_MAP[theme]
        elif theme not in VALID_THEME_SET:
            raise ValueError(
                f"Invalid theme name, theme must be one of: {', '.join(VALID_THEME_SET)}"
            )
        self.theme = theme

        if css is not None and os.path.exists(css):
            with open(css) as css_file:
                self.css = css_file.read()
        else:
            self.css = css
        if examples is None or (
            isinstance(examples, list)
            and (len(examples) == 0 or isinstance(examples[0], list))
        ):
            self.examples = examples
        elif (
            isinstance(examples, list) and len(self.input_components) == 1
        ):  # If there is only one input component, examples can be provided as a regular list instead of a list of lists
            self.examples = [[e] for e in examples]
        elif isinstance(examples, str):
            if not os.path.exists(examples):
                raise FileNotFoundError(
                    "Could not find examples directory: " + examples
                )
            log_file = os.path.join(examples, "log.csv")
            if not os.path.exists(log_file):
                if len(self.input_components) == 1:
                    exampleset = [
                        [os.path.join(examples, item)] for item in os.listdir(examples)
                    ]
                else:
                    raise FileNotFoundError(
                        "Could not find log file (required for multiple inputs): "
                        + log_file
                    )
            else:
                with open(log_file) as logs:
                    exampleset = list(csv.reader(logs))
                    exampleset = exampleset[1:]  # remove header
            for i, example in enumerate(exampleset):
                for j, (component, cell) in enumerate(
                    zip(
                        self.input_components + self.output_components,
                        example,
                    )
                ):
                    exampleset[i][j] = component.restore_flagged(
                        examples,
                        cell,
                        None,
                    )
            self.examples = exampleset
        else:
            raise ValueError(
                "Examples argument must either be a directory or a nested "
                "list, where each sublist represents a set of inputs."
            )
        self.num_shap = num_shap
        self.examples_per_page = examples_per_page

        self.simple_server = None

        # For analytics_enabled and allow_flagging: (1) first check for
        # parameter, (2) check for env variable, (3) default to True/"manual"
        self.analytics_enabled = (
            analytics_enabled
            if analytics_enabled is not None
            else os.getenv("GRADIO_ANALYTICS_ENABLED", "True") == "True"
        )
        if allow_flagging is None:
            allow_flagging = os.getenv("GRADIO_ALLOW_FLAGGING", "manual")
        if allow_flagging is True:
            warnings.warn(
                "The `allow_flagging` parameter in `Interface` now"
                "takes a string value ('auto', 'manual', or 'never')"
                ", not a boolean. Setting parameter to: 'manual'."
            )
            self.allow_flagging = "manual"
        elif allow_flagging == "manual":
            self.allow_flagging = "manual"
        elif allow_flagging is False:
            warnings.warn(
                "The `allow_flagging` parameter in `Interface` now"
                "takes a string value ('auto', 'manual', or 'never')"
                ", not a boolean. Setting parameter to: 'never'."
            )
            self.allow_flagging = "never"
        elif allow_flagging == "never":
            self.allow_flagging = "never"
        elif allow_flagging == "auto":
            self.allow_flagging = "auto"
        else:
            raise ValueError(
                "Invalid value for `allow_flagging` parameter."
                "Must be: 'auto', 'manual', or 'never'."
            )

        self.flagging_options = flagging_options
        self.flagging_callback = flagging_callback
        self.flagging_dir = flagging_dir

        self.save_to = None  # Used for selenium tests
        self.share = None
        self.share_url = None
        self.local_url = None

        self.requires_permissions = any(
            [component.requires_permissions for component in self.input_components]
        )

        self.favicon_path = None

        data = {
            "fn": fn,
            "inputs": inputs,
            "outputs": outputs,
            "live": live,
            "ip_address": self.ip_address,
            "interpretation": interpretation,
            "allow_flagging": allow_flagging,
            "custom_css": self.css is not None,
            "theme": self.theme,
        }

        if self.analytics_enabled:
            utils.initiated_analytics(data)

        utils.version_check()
        Interface.instances.add(self)

        param_names = inspect.getfullargspec(self.predict[0])[0]
        for component, param_name in zip(self.input_components, param_names):
            if component.label is None:
                component.label = param_name
        for i, component in enumerate(self.output_components):
            if component.label is None:
                if len(self.output_components) == 1:
                    component.label = "output"
                else:
                    component.label = "output " + str(i)

        if self.cache_examples and examples:
            cache_interface_examples(self)

        if self.allow_flagging != "never":
            self.flagging_callback.setup(
                self.input_components + self.output_components, self.flagging_dir
            )

        with self:
            if self.title:
                Markdown(
                    "<h1 style='text-align: center; margin-bottom: 1rem'>"
                    + self.title
                    + "</h1>"
                )
            if self.description:
                Markdown(self.description)
            with Row():
                if self.interface_type in [
                    self.InterfaceTypes.STANDARD,
                    self.InterfaceTypes.INPUT_ONLY,
                    self.InterfaceTypes.UNIFIED,
                ]:
                    with Column(variant="panel"):
                        input_component_column = Column()
                        if self.interface_type in [
                            self.InterfaceTypes.INPUT_ONLY,
                            self.InterfaceTypes.UNIFIED,
                        ]:
                            status_tracker = StatusTracker(cover_container=True)
                        with input_component_column:
                            for component in self.input_components:
                                component.render()
                        if self.interpretation:
                            interpret_component_column = Column(visible=False)
                            interpretation_set = []
                            with interpret_component_column:
                                for component in self.input_components:
                                    interpretation_set.append(Interpretation(component))
                        with Row():
                            if self.interface_type in [
                                self.InterfaceTypes.STANDARD,
                                self.InterfaceTypes.INPUT_ONLY,
                            ]:
                                clear_btn = Button("Clear")
                                if not self.live:
                                    submit_btn = Button("Submit")
                            elif self.interface_type == self.InterfaceTypes.UNIFIED:
                                clear_btn = Button("Clear")
                                submit_btn = Button("Submit")
                                if self.allow_flagging == "manual":
                                    flag_btn = Button("Flag")
                                    flag_btn.click(
                                        lambda *flag_data: self.flagging_callback.flag(
                                            flag_data
                                        ),
                                        inputs=self.input_components,
                                        outputs=[],
                                        _preprocess=False,
                                        queue=False,
                                    )

                if self.interface_type in [
                    self.InterfaceTypes.STANDARD,
                    self.InterfaceTypes.OUTPUT_ONLY,
                ]:

                    with Column(variant="panel"):
                        status_tracker = StatusTracker(cover_container=True)
                        for component in self.output_components:
                            component.render()
                        with Row():
                            if self.interface_type == self.InterfaceTypes.OUTPUT_ONLY:
                                clear_btn = Button("Clear")
                                submit_btn = Button("Generate")
                            if self.allow_flagging == "manual":
                                flag_btn = Button("Flag")
                                flag_btn.click(
                                    lambda *flag_data: self.flagging_callback.flag(
                                        flag_data
                                    ),
                                    inputs=self.input_components
                                    + self.output_components,
                                    outputs=[],
                                    _preprocess=False,
                                )
                            if self.interpretation:
                                interpretation_btn = Button("Interpret")
            submit_fn = (
                lambda *args: self.run_prediction(args)[0]
                if len(self.output_components) == 1
                else self.run_prediction(args)
            )
            if self.live:
                for component in self.input_components:
                    component.change(
                        submit_fn, self.input_components, self.output_components
                    )
            else:
                submit_btn.click(
                    submit_fn,
                    self.input_components,
                    self.output_components,
                    status_tracker=status_tracker,
                )
            clear_btn.click(
                None,
                [],
                (
                    self.input_components
                    + self.output_components
                    + (
                        [input_component_column]
                        if self.interface_type
                        in [
                            self.InterfaceTypes.STANDARD,
                            self.InterfaceTypes.INPUT_ONLY,
                            self.InterfaceTypes.UNIFIED,
                        ]
                        else []
                    )
                    + ([interpret_component_column] if self.interpretation else [])
                ),
                _js=f"""() => {json.dumps(
                    [component.cleared_value if hasattr(component, "cleared_value") else None
                    for component in self.input_components + self.output_components] + (
                            [True]
                            if self.interface_type
                            in [
                                self.InterfaceTypes.STANDARD,
                                self.InterfaceTypes.INPUT_ONLY,
                                self.InterfaceTypes.UNIFIED,
                            ]
                            else []
                        )
                    + ([False] if self.interpretation else [])
                )}
                """,
            )
            if self.examples:
                non_state_inputs = [
                    c for c in self.input_components if not isinstance(c, Variable)
                ]

                examples = Dataset(
                    components=non_state_inputs,
                    samples=self.examples,
                    type="index",
                )

                def load_example(example_id):
                    processed_examples = [
                        component.preprocess_example(sample)
                        for component, sample in zip(
                            self.input_components, self.examples[example_id]
                        )
                    ]
                    if self.cache_examples:
                        processed_examples += load_from_cache(self, example_id)
                    if len(processed_examples) == 1:
                        return processed_examples[0]
                    else:
                        return processed_examples

                examples.click(
                    load_example,
                    inputs=[examples],
                    outputs=non_state_inputs
                    + (self.output_components if self.cache_examples else []),
                    _postprocess=False,
                    queue=False,
                )

            if self.interpretation:
                interpretation_btn.click(
                    lambda *data: self.interpret(data) + [False, True],
                    inputs=self.input_components + self.output_components,
                    outputs=interpretation_set
                    + [input_component_column, interpret_component_column],
                    status_tracker=status_tracker,
                    _preprocess=False,
                )

            if self.article:
                Markdown(self.article)

        self.config = self.get_config_file()

    def __call__(self, *params):
        if (
            self.api_mode
        ):  # skip the preprocessing/postprocessing if sending to a remote API
            output = self.run_prediction(params, called_directly=True)
        else:
            output = self.process(params)
        return output[0] if len(output) == 1 else output

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        repr = "Gradio Interface for: {}".format(
            ", ".join(fn.__name__ for fn in self.predict)
        )
        repr += "\n" + "-" * len(repr)
        repr += "\ninputs:"
        for component in self.input_components:
            repr += "\n|-{}".format(str(component))
        repr += "\noutputs:"
        for component in self.output_components:
            repr += "\n|-{}".format(str(component))
        return repr

    def run_prediction(
        self,
        processed_input: List[Any],
        called_directly: bool = False,
    ) -> List[Any] | Tuple[List[Any], List[float]]:
        """
        Runs the prediction function with the given (already processed) inputs.
        Parameters:
        processed_input (list): A list of processed inputs.
        called_directly (bool): Whether the prediction is being called
            directly (i.e. as a function, not through the GUI).
        Returns:
        predictions (list): A list of predictions (not post-processed).
        """
        if self.api_mode:  # Serialize the input
            processed_input = [
                input_component.serialize(processed_input[i], called_directly)
                for i, input_component in enumerate(self.input_components)
            ]
        predictions = []
        output_component_counter = 0

        for predict_fn in self.predict:
            prediction = predict_fn(*processed_input)

            if len(self.output_components) == len(self.predict) or prediction is None:
                prediction = [prediction]

            if self.api_mode:  # Serialize the input
                prediction_ = copy.deepcopy(prediction)
                prediction = []

                # Done this way to handle both single interfaces with multiple outputs and Parallel() interfaces
                for pred in prediction_:
                    prediction.append(
                        self.output_components[output_component_counter].deserialize(
                            pred
                        )
                    )
                    output_component_counter += 1

            predictions.extend(prediction)

        return predictions

    def process(self, raw_input: List[Any]) -> Tuple[List[Any], List[float]]:
        """
        First preprocesses the input, then runs prediction using
        self.run_prediction(), then postprocesses the output.
        Parameters:
        raw_input: a list of raw inputs to process and apply the prediction(s) on.
        Returns:
        processed output: a list of processed  outputs to return as the prediction(s).
        duration: a list of time deltas measuring inference time for each prediction fn.
        """
        processed_input = [
            input_component.preprocess(raw_input[i])
            for i, input_component in enumerate(self.input_components)
        ]
        predictions = self.run_prediction(processed_input)
        processed_output = [
            output_component.postprocess(predictions[i])
            if predictions[i] is not None
            else None
            for i, output_component in enumerate(self.output_components)
        ]
        return processed_output

    def interpret(self, raw_input: List[Any]) -> List[Any]:
        return [
            {"original": raw_value, "interpretation": interpretation}
            for interpretation, raw_value in zip(
                interpretation.run_interpret(self, raw_input)[0], raw_input
            )
        ]

    def test_launch(self) -> None:
        """
        Passes a few samples through the function to test if the inputs/outputs
        components are consistent with the function parameter and return values.
        """
        for predict_fn in self.predict:
            print("Test launch: {}()...".format(predict_fn.__name__), end=" ")
            raw_input = []
            for input_component in self.input_components:
                if input_component.test_input is None:
                    print("SKIPPED")
                    break
                else:
                    raw_input.append(input_component.test_input)
            else:
                self.process(raw_input)
                print("PASSED")
                continue

    def integrate(self, comet_ml=None, wandb=None, mlflow=None) -> None:
        """
        A catch-all method for integrating with other libraries.
        Should be run after launch()
        Parameters:
            comet_ml (Experiment): If a comet_ml Experiment object is provided,
            will integrate with the experiment and appear on Comet dashboard
            wandb (module): If the wandb module is provided, will integrate
            with it and appear on WandB dashboard
            mlflow (module): If the mlflow module  is provided, will integrate
            with the experiment and appear on ML Flow dashboard
        """
        analytics_integration = ""
        if comet_ml is not None:
            analytics_integration = "CometML"
            comet_ml.log_other("Created from", "Gradio")
            if self.share_url is not None:
                comet_ml.log_text("gradio: " + self.share_url)
                comet_ml.end()
            else:
                comet_ml.log_text("gradio: " + self.local_url)
                comet_ml.end()
        if wandb is not None:
            analytics_integration = "WandB"
            if self.share_url is not None:
                wandb.log(
                    {
                        "Gradio panel": wandb.Html(
                            '<iframe src="'
                            + self.share_url
                            + '" width="'
                            + str(self.width)
                            + '" height="'
                            + str(self.height)
                            + '" frameBorder="0"></iframe>'
                        )
                    }
                )
            else:
                print(
                    "The WandB integration requires you to "
                    "`launch(share=True)` first."
                )
        if mlflow is not None:
            analytics_integration = "MLFlow"
            if self.share_url is not None:
                mlflow.log_param("Gradio Interface Share Link", self.share_url)
            else:
                mlflow.log_param("Gradio Interface Local Link", self.local_url)
        if self.analytics_enabled and analytics_integration:
            data = {"integration": analytics_integration}
            utils.integration_analytics(data)


class TabbedInterface(Blocks):
    def __init__(
        self, interface_list: List[Interface], tab_names: Optional[List[str]] = None
    ):
        if tab_names is None:
            tab_names = ["Tab {}".format(i) for i in range(len(interface_list))]
        super().__init__()
        with self:
            with Tabs():
                for (interface, tab_name) in zip(interface_list, tab_names):
                    with TabItem(label=tab_name):
                        interface.render()


def close_all(verbose: bool = True) -> None:
    for io in Interface.get_instances():
        io.close(verbose)



from __future__ import annotations

import inspect
import json
import math
import numbers
import operator
import os
import shutil
import sys
import tempfile
import warnings
from copy import deepcopy
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple

import matplotlib.figure
import numpy as np
import pandas as pd
import PIL
from ffmpy import FFmpeg
from markdown_it import MarkdownIt

from gradio import media_data, processing_utils
from gradio.blocks import Block
from gradio.css import CSS, CSSWithLabel

class Component(Block):
    """
    A base class for defining the methods that all gradio components should have.
    """

    def __init__(
        self,
        *,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        super().__init__(css=css, **kwargs)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.get_block_name()} (label={self.label})"

    def get_template_context(self):
        """
        :return: a dictionary with context variables for the javascript file associated with the context
        """
        return {
            "name": self.get_block_name(),
            **super().get_template_context()
        }

    @staticmethod
    def update(**kwargs) -> dict:
        """
        Updates component parameters

        @param kwargs: Updating component parameters
        @return: Updated component parameters
        """
        kwargs["__type__"] = "update"
        return kwargs


class IOComponent(Component):
    """
    A base class for defining methods that all input/output components should have.
    """

    def __init__(
        self,
        *,
        label: Optional[str] = None,
        show_label: bool = True,
        requires_permissions: bool = False,
        interactive: Optional[bool] = None,
        **kwargs,
    ):
        self.label = label
        self.show_label = show_label
        self.requires_permissions = requires_permissions
        self.interactive = interactive

        self.set_interpret_parameters()

        super().__init__(**kwargs)

    def get_template_context(self):
        return {
            "label": self.label,
            "show_label": self.show_label,
            "interactive": self.interactive,
            **super().get_template_context(),
        }

    def save_flagged(
        self, dir: str, label: Optional[str], data: Any, encryption_key: bool
    ) -> Any:
        """
        Saves flagged data from component
        """
        return data

    def restore_flagged(self, dir, data, encryption_key):
        """
        Restores flagged data from logs
        """
        return data

    def save_file(self, file: tempfile._TemporaryFileWrapper, dir: str, label: str):
        """
        Saved flagged file and returns filepath
        """
        label = "".join([char for char in label if char.isalnum() or char in "._- "])
        old_file_name = file.name
        output_dir = os.path.join(dir, label)
        if os.path.exists(output_dir):
            file_index = len(os.listdir(output_dir))
        else:
            os.makedirs(output_dir)
            file_index = 0
        new_file_name = str(file_index)
        if "." in old_file_name:
            uploaded_format = old_file_name.split(".")[-1].lower()
            new_file_name += "." + uploaded_format
        file.close()
        shutil.move(old_file_name, os.path.join(dir, label, new_file_name))
        return label + "/" + new_file_name

    def save_flagged_file(
        self,
        dir: str,
        label: str,
        data: Any,
        encryption_key: bool,
        file_path: Optional[str] = None,
    ) -> Optional[str]:
        """
        Saved flagged data (e.g. image or audio) as a file and returns filepath
        """
        if data is None:
            return None
        file = processing_utils.decode_base64_to_file(data, encryption_key, file_path)
        return self.save_file(file, dir, label)

    def restore_flagged_file(
        self,
        dir: str,
        file: str,
        encryption_key: bool,
    ) -> Dict[str, Any]:
        """
        Loads flagged data from file and returns it
        """
        data = processing_utils.encode_file_to_base64(
            os.path.join(dir, file), encryption_key=encryption_key
        )
        return {"name": file, "data": data}

    # Input Functionalities
    def preprocess(self, x: Any) -> Any:
        """
        Any preprocessing needed to be performed on function input.
        """
        return x

    def serialize(self, x: Any, called_directly: bool) -> Any:
        """
        Convert from a human-readable version of the input (path of an image, URL of a video, etc.) into the interface to a serialized version (e.g. base64) to pass into an API. May do different things if the interface is called() vs. used via GUI.
        Parameters:
        x (Any): Input to interface
        called_directly (bool): if true, the interface was called(), otherwise, it is being used via the GUI
        """
        return x

    def preprocess_example(self, x: Any) -> Any:
        """
        Any preprocessing needed to be performed on an example before being passed to the main function.
        """
        return x

    def set_interpret_parameters(self):
        """
        Set any parameters for interpretation.
        """
        return self

    def get_interpretation_neighbors(self, x: Any) -> Tuple[List[Any], Dict[Any], bool]:
        """
        Generates values similar to input to be used to interpret the significance of the input in the final output.
        Parameters:
        x (Any): Input to interface
        Returns: (neighbor_values, interpret_kwargs, interpret_by_removal)
        neighbor_values (List[Any]): Neighboring values to input x to compute for interpretation
        interpret_kwargs (Dict[Any]): Keyword arguments to be passed to get_interpretation_scores
        interpret_by_removal (bool): If True, returned neighbors are values where the interpreted subsection was removed. If False, returned neighbors are values where the interpreted subsection was modified to a different value.
        """
        return [], {}, True

    def get_interpretation_scores(
        self, x: Any, neighbors: List[Any], scores: List[float], **kwargs
    ) -> List[Any]:
        """
        Arrange the output values from the neighbors into interpretation scores for the interface to render.
        Parameters:
        x (Any): Input to interface
        neighbors (List[Any]): Neighboring values to input x used for interpretation.
        scores (List[float]): Output value corresponding to each neighbor in neighbors
        kwargs (Dict[str, Any]): Any additional arguments passed from get_interpretation_neighbors.
        Returns:
        (List[Any]): Arrangement of interpretation scores for interfaces to render.
        """
        pass

    def generate_sample(self) -> Any:
        """
        Returns a sample value of the input that would be accepted by the api. Used for api documentation.
        """
        pass

    # Output Functionalities
    def postprocess(self, y):
        """
        Any postprocessing needed to be performed on function output.
        """
        return y

    def deserialize(self, x):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return x


class Changeable(Component):
    def change(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        status_tracker: Optional[StatusTracker] = None,
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            status_tracker: StatusTracker to visualize function progress
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of input and outputs components, return should be a list of values for output component.
        Returns: None
        """
        self.set_event_trigger(
            "change", fn, inputs, outputs, status_tracker=status_tracker, js=_js
        )


class Clickable(Component):
    def click(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        status_tracker: Optional[StatusTracker] = None,
        queue=None,
        _js: Optional[str] = None,
        _preprocess: bool = True,
        _postprocess: bool = True,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            status_tracker: StatusTracker to visualize function progress
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            _preprocess: If False, will not run preprocessing of component data before running 'fn'.
            _postprocess: If False, will not run postprocessing of component data before returning 'fn' output.
        Returns: None
        """
        self.set_event_trigger(
            "click",
            fn,
            inputs,
            outputs,
            status_tracker=status_tracker,
            queue=queue,
            js=_js,
            preprocess=_preprocess,
            postprocess=_postprocess,
        )


class Submittable(Component):
    def submit(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        status_tracker: Optional[StatusTracker] = None,
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            status_tracker: StatusTracker to visualize function progress
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger(
            "submit", fn, inputs, outputs, status_tracker=status_tracker, js=_js
        )


class Editable(Component):
    def edit(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger("edit", fn, inputs, outputs, js=_js)


class Clearable(Component):
    def clear(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger("submit", fn, inputs, outputs, js=_js)


class Playable(Component):
    def play(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger("play", fn, inputs, outputs, js=_js)

    def pause(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger("pause", fn, inputs, outputs, js=_js)

    def stop(
        self,
        fn: Callable,
        inputs: List[Component],
        outputs: List[Component],
        _js: Optional[str] = None,
    ):
        """
        Parameters:
            fn: Callable function
            inputs: List of inputs
            outputs: List of outputs
            _js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
        Returns: None
        """
        self.set_event_trigger("stop", fn, inputs, outputs, js=_js)


class Textbox(Changeable, Submittable, IOComponent):
    """
    Component creates a textbox for user to enter string input or display string output. Provides a string as an argument to the wrapped function.
    Input type: str
    Output type: str

    Demos: hello_world, diff_texts, sentence_builder
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        lines: int = 1,
        max_lines: int = 20,
        placeholder: Optional[str] = None,
        label: Optional[str] = None,
        css: Optional[CSSWithLabel] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): default text to provide in textarea.
        lines (int): minimum number of line rows to provide in textarea.
        max_lines (int): maximum number of line rows to provide in textarea.
        placeholder (str): placeholder hint to provide behind textarea.
        label (str): component name in interface.
        """
        default_value = str(default_value)
        self.lines = lines
        self.max_lines = max_lines
        self.placeholder = placeholder
        self.default_value = default_value
        self.cleared_value = ""
        self.test_input = default_value
        self.interpret_by_tokens = True
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "lines": self.lines,
            "max_lines": self.max_lines,
            "placeholder": self.placeholder,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    # Input Functionalities
    def preprocess(self, x: str | None) -> Any:
        """
        Any preprocessing needed to be performed on function input.
        """
        if x is None:
            return None
        else:
            return str(x)

    def serialize(self, x: Any, called_directly: bool) -> Any:
        """
        Convert from a human-readable version of the input (path of an image, URL of a video, etc.) into the interface to a serialized version (e.g. base64) to pass into an API. May do different things if the interface is called() vs. used via GUI.
        Parameters:
        x (Any): Input to interface
        called_directly (bool): if true, the interface was called(), otherwise, it is being used via the GUI
        """
        return x

    def preprocess_example(self, x: str | None) -> Any:
        """
        Any preprocessing needed to be performed on an example before being passed to the main function.
        """
        if x is None:
            return None
        else:
            return str(x)

    def set_interpret_parameters(
        self, separator: str = " ", replacement: Optional[str] = None
    ):
        """
        Calculates interpretation score of characters in input by splitting input into tokens, then using a "leave one out" method to calculate the score of each token by removing each token and measuring the delta of the output value.
        Parameters:
        separator (str): Separator to use to split input into tokens.
        replacement (str): In the "leave one out" step, the text that the token should be replaced with. If None, the token is removed altogether.
        """
        self.interpretation_separator = separator
        self.interpretation_replacement = replacement
        return self

    def tokenize(self, x: str) -> Tuple[List[str], List[str], None]:
        """
        Tokenizes an input string by dividing into "words" delimited by self.interpretation_separator
        """
        tokens = x.split(self.interpretation_separator)
        leave_one_out_strings = []
        for index in range(len(tokens)):
            leave_one_out_set = list(tokens)
            if self.interpretation_replacement is None:
                leave_one_out_set.pop(index)
            else:
                leave_one_out_set[index] = self.interpretation_replacement
            leave_one_out_strings.append(
                self.interpretation_separator.join(leave_one_out_set)
            )
        return tokens, leave_one_out_strings, None

    def get_masked_inputs(
        self, tokens: List[str], binary_mask_matrix: List[List[int]]
    ) -> List[str]:
        """
        Constructs partially-masked sentences for SHAP interpretation
        """
        masked_inputs = []
        for binary_mask_vector in binary_mask_matrix:
            masked_input = np.array(tokens)[np.array(binary_mask_vector, dtype=bool)]
            masked_inputs.append(self.interpretation_separator.join(masked_input))
        return masked_inputs

    def get_interpretation_scores(
        self, x, neighbors, scores: List[float], tokens: List[str], masks=None, **kwargs
    ) -> List[Tuple[str, float]]:
        """
        Returns:
        (List[Tuple[str, float]]): Each tuple set represents a set of characters and their corresponding interpretation score.
        """
        result = []
        for token, score in zip(tokens, scores):
            result.append((token, score))
            result.append((self.interpretation_separator, 0))
        return result

    def generate_sample(self) -> str:
        return "Hello World"

    # Output Functionalities
    def postprocess(self, y: str | None):
        """
        Any postprocessing needed to be performed on function output.
        """
        if y is None:
            return None
        else:
            return str(y)

    def deserialize(self, x):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return x


class Number(Changeable, Submittable, IOComponent):
    """
    Component creates a field for user to enter numeric input or display numeric output. Provides a number as an argument to the wrapped function.
    Can be used as an output as well.

    Input type: float
    Output type: float
    Demos: tax_calculator, titanic_survival
    """

    def __init__(
        self,
        default_value: Optional[float] = None,
        *,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (float): default value.
        label (str): component name in interface.
        """
        self.default_value = float(default_value) if default_value is not None else None
        self.test_input = self.default_value if self.default_value is not None else 1
        self.interpret_by_tokens = False
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: float | None) -> Optional[float]:
        """
        Parameters:
        x (string): numeric input as a string
        Returns:
        (float): number representing function input
        """
        if x is None:
            return None
        return float(x)

    def preprocess_example(self, x: float | None) -> float | None:
        """
        Returns:
        (float): Number representing function input
        """
        if x is None:
            return None
        else:
            return float(x)

    def set_interpret_parameters(
        self, steps: int = 3, delta: float = 1, delta_type: str = "percent"
    ):
        """
        Calculates interpretation scores of numeric values close to the input number.
        Parameters:
        steps (int): Number of nearby values to measure in each direction (above and below the input number).
        delta (float): Size of step in each direction between nearby values.
        delta_type (str): "percent" if delta step between nearby values should be a calculated as a percent, or "absolute" if delta should be a constant step change.
        """
        self.interpretation_steps = steps
        self.interpretation_delta = delta
        self.interpretation_delta_type = delta_type
        return self

    def get_interpretation_neighbors(self, x: float) -> Tuple[List[float], Dict]:
        x = float(x)
        if self.interpretation_delta_type == "percent":
            delta = 1.0 * self.interpretation_delta * x / 100
        elif self.interpretation_delta_type == "absolute":
            delta = self.interpretation_delta
        else:
            delta = self.interpretation_delta
        negatives = (x + np.arange(-self.interpretation_steps, 0) * delta).tolist()
        positives = (x + np.arange(1, self.interpretation_steps + 1) * delta).tolist()
        return negatives + positives, {}

    def get_interpretation_scores(
        self, x: Number, neighbors: List[float], scores: List[float], **kwargs
    ) -> List[Tuple[float, float]]:
        """
        Returns:
        (List[Tuple[float, float]]): Each tuple set represents a numeric value near the input and its corresponding interpretation score.
        """
        interpretation = list(zip(neighbors, scores))
        interpretation.insert(int(len(interpretation) / 2), [x, None])
        return interpretation

    def generate_sample(self) -> float:
        return 1.0

    # Output Functionalities
    def postprocess(self, y: float | None):
        """
        Any postprocessing needed to be performed on function output.
        """
        if y is None:
            return None
        else:
            return float(y)

    def deserialize(self, y):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return y


class Slider(Changeable, IOComponent):
    """
    Component creates a slider that ranges from `minimum` to `maximum`. Provides a number as an argument to the wrapped function.

    Input type: float
    Demos: sentence_builder, generate_tone, titanic_survival
    """

    def __init__(
        self,
        default_value: Optional[float] = None,
        *,
        minimum: float = 0,
        maximum: float = 100,
        step: Optional[float] = None,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (float): default value.
        minimum (float): minimum value for slider.
        maximum (float): maximum value for slider.
        step (float): increment between slider values.
        label (str): component name in interface.
        """
        self.minimum = minimum
        self.maximum = maximum
        if step is None:
            difference = maximum - minimum
            power = math.floor(math.log10(difference) - 2)
            step = 10**power
        self.step = step
        self.default_value = minimum if default_value is None else default_value
        self.cleared_value = self.default_value
        self.test_input = self.default_value
        self.interpret_by_tokens = False
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "minimum": self.minimum,
            "maximum": self.maximum,
            "step": self.step,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: float) -> float:
        """
        Parameters:
        x (number): numeric input
        Returns:
        (number): numeric input
        """
        return x

    def preprocess_example(self, x: float) -> float:
        """
        Returns:
        (float): Number representing function input
        """
        return x

    def set_interpret_parameters(self, steps: int = 8) -> "Slider":
        """
        Calculates interpretation scores of numeric values ranging between the minimum and maximum values of the slider.
        Parameters:
        steps (int): Number of neighboring values to measure between the minimum and maximum values of the slider range.
        """
        self.interpretation_steps = steps
        return self

    def get_interpretation_neighbors(self, x) -> Tuple[object, dict]:
        return (
            np.linspace(self.minimum, self.maximum, self.interpretation_steps).tolist(),
            {},
        )

    def get_interpretation_scores(
        self, x, neighbors, scores: List[float], **kwargs
    ) -> List[float]:
        """
        Returns:
        (List[float]): Each value represents the score corresponding to an evenly spaced range of inputs between the minimum and maximum slider values.
        """
        return scores

    def generate_sample(self) -> float:
        return self.maximum

        # Output Functionalities

    def postprocess(self, y: float | None):
        """
        Any postprocessing needed to be performed on function output.
        """
        return y

    def deserialize(self, y):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return y


class Checkbox(Changeable, IOComponent):
    """
    Component creates a checkbox that can be set to `True` or `False`. Provides a boolean as an argument to the wrapped function.

    Input type: bool
    Output type: bool
    Demos: sentence_builder, titanic_survival
    """

    def __init__(
        self,
        default_value: bool = False,
        *,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (bool): if True, checked by default.
        label (str): component name in interface.
        """
        self.test_input = True
        self.default_value = default_value
        self.interpret_by_tokens = False
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: bool) -> bool:
        """
        Parameters:
        x (bool): boolean input
        Returns:
        (bool): boolean input
        """
        return x

    def preprocess_example(self, x):
        """
        Returns:
        (bool): Boolean representing function input
        """
        return x

    def set_interpret_parameters(self):
        """
        Calculates interpretation score of the input by comparing the output against the output when the input is the inverse boolean value of x.
        """
        return self

    def get_interpretation_neighbors(self, x):
        return [not x], {}

    def get_interpretation_scores(self, x, neighbors, scores, **kwargs):
        """
        Returns:
        (Tuple[float, float]): The first value represents the interpretation score if the input is False, and the second if the input is True.
        """
        if x:
            return scores[0], None
        else:
            return None, scores[0]

    def generate_sample(self):
        return True

    # Output Functionalities
    def postprocess(self, y):
        """
        Any postprocessing needed to be performed on function output.
        """
        return y

    def deserialize(self, x):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return x


class CheckboxGroup(Changeable, IOComponent):
    """
    Component creates a set of checkboxes of which a subset can be selected. Provides a list of strings representing the selected choices as an argument to the wrapped function.

    Input type: Union[List[str], List[int]]
    Demos: sentence_builder, titanic_survival, fraud_detector
    """

    def __init__(
        self,
        choices: List[str],
        *,
        default_selected: List[str] = None,
        type: str = "value",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        choices (List[str]): list of options to select from.
        default_selected (List[str]): default selected list of options.
        type (str): Type of value to be returned by component. "value" returns the list of strings of the choices selected, "index" returns the list of indicies of the choices selected.
        label (str): component name in interface.
        """
        if (
            default_selected is None
        ):  # Mutable parameters shall not be given as default parameters in the function.
            default_selected = []
        self.choices = choices
        self.default_value = default_selected
        self.cleared_value = []
        self.type = type
        self.test_input = self.choices
        self.interpret_by_tokens = False
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "choices": self.choices,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: List[str]) -> List[str] | List[int]:
        """
        Parameters:
        x (List[str]): list of selected choices
        Returns:
        (Union[List[str], List[int]]): list of selected choices as strings or indices within choice list
        """
        if self.type == "value":
            return x
        elif self.type == "index":
            return [self.choices.index(choice) for choice in x]
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'value', 'index'."
            )

    def set_interpret_parameters(self):
        """
        Calculates interpretation score of each choice in the input by comparing the output against the outputs when each choice in the input is independently either removed or added.
        """
        return self

    def get_interpretation_neighbors(self, x):
        leave_one_out_sets = []
        for choice in self.choices:
            leave_one_out_set = list(x)
            if choice in leave_one_out_set:
                leave_one_out_set.remove(choice)
            else:
                leave_one_out_set.append(choice)
            leave_one_out_sets.append(leave_one_out_set)
        return leave_one_out_sets, {}

    def get_interpretation_scores(self, x, neighbors, scores, **kwargs):
        """
        Returns:
        (List[Tuple[float, float]]): For each tuple in the list, the first value represents the interpretation score if the input is False, and the second if the input is True.
        """
        final_scores = []
        for choice, score in zip(self.choices, scores):
            if choice in x:
                score_set = [score, None]
            else:
                score_set = [None, score]
            final_scores.append(score_set)
        return final_scores

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (List[str]])
        """
        return json.dumps(data)

    def restore_flagged(self, dir, data, encryption_key):
        return json.loads(data)

    def generate_sample(self):
        return self.choices

    # Output Functionalities
    def postprocess(self, y):
        """
        Any postprocessing needed to be performed on function output.
        """
        return y

    def deserialize(self, x):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return x


class Radio(Changeable, IOComponent):
    """
    Component creates a set of radio buttons of which only one can be selected. Provides string representing selected choice as an argument to the wrapped function.

    Input type: Union[str, int]
    Demos: sentence_builder, tax_calculator, titanic_survival
    """

    def __init__(
        self,
        choices: List[str],
        *,
        default_selected: Optional[str] = None,
        type: str = "value",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        choices (List[str]): list of options to select from.
        default_selected (str): the button selected by default. If None, no button is selected by default.
        type (str): Type of value to be returned by component. "value" returns the string of the choice selected, "index" returns the index of the choice selected.
        label (str): component name in interface.
        """
        self.choices = choices
        self.type = type
        self.test_input = self.choices[0]
        self.default_value = (
            default_selected if default_selected is not None else self.choices[0]
        )
        self.cleared_value = self.default_value
        self.interpret_by_tokens = False
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "choices": self.choices,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: str) -> str | int:
        """
        Parameters:
        x (str): selected choice
        Returns:
        (Union[str, int]): selected choice as string or index within choice list
        """
        if self.type == "value":
            return x
        elif self.type == "index":
            if x is None:
                return None
            else:
                return self.choices.index(x)
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'value', 'index'."
            )

    def set_interpret_parameters(self):
        """
        Calculates interpretation score of each choice by comparing the output against each of the outputs when alternative choices are selected.
        """
        return self

    def get_interpretation_neighbors(self, x):
        choices = list(self.choices)
        choices.remove(x)
        return choices, {}

    def get_interpretation_scores(self, x, neighbors, scores, **kwargs):
        """
        Returns:
        (List[float]): Each value represents the interpretation score corresponding to each choice.
        """
        scores.insert(self.choices.index(x), None)
        return scores

    def generate_sample(self):
        return self.choices[0]

    # Output Functionalities
    def postprocess(self, y):
        """
        Any postprocessing needed to be performed on function output.
        """
        return y

    def deserialize(self, x):
        """
        Convert from serialized output (e.g. base64 representation) from a call() to the interface to a human-readable version of the output (path of an image, etc.)
        """
        return x


class Dropdown(Radio):
    """
    Component creates a dropdown of which only one can be selected. Provides string representing selected choice as an argument to the wrapped function.

    Input type: Union[str, int]
    Demos: sentence_builder, filter_records, titanic_survival
    """

    def __init__(
        self,
        choices: List[str],
        *,
        default_selected: Optional[str] = None,
        type: str = "value",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        choices (List[str]): list of options to select from.
        default_selected (str): default value selected in dropdown. If None, no value is selected by default.
        type (str): Type of value to be returned by component. "value" returns the string of the choice selected, "index" returns the index of the choice selected.
        label (str): component name in interface.
        """
        # Everything is same with Dropdown and Radio, so let's make use of it :)
        IOComponent.__init__(
            self,
            default_selected=default_selected,
            choices=choices,
            type=type,
            label=label,
            **kwargs,
        )


class Image(Editable, Clearable, IOComponent):
    """
    Component creates an image component with input and output capabilities.

    Input type: Union[numpy.array, PIL.Image, file-object]
    Output type: Union[numpy.array, PIL.Image, str, matplotlib.pyplot, Tuple[Union[numpy.array, PIL.Image, str], List[Tuple[str, float, float, float, float]]]]
    Demos: image_classifier, image_mod, webcam, digit_classifier
    """

    def __init__(
        self,
        default_value: Optional[str] = None,
        *,
        shape: Tuple[int, int] = None,
        image_mode: str = "RGB",
        invert_colors: bool = False,
        source: str = "upload",
        tool: str = "editor",
        type: str = "numpy",
        label: str = None,
        css: Optional[CSSWithLabel] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value(str): A path or URL for the default value that Image component is going to take.
        shape (Tuple[int, int]): (width, height) shape to crop and resize image to; if None, matches input image size. Pass None for either width or height to only crop and resize the other.
        image_mode (str): "RGB" if color, or "L" if black and white.
        invert_colors (bool): whether to invert the image as a preprocessing step.
        source (str): Source of image. "upload" creates a box where user can drop an image file, "webcam" allows user to take snapshot from their webcam, "canvas" defaults to a white image that can be edited and drawn upon with tools.
        tool (str): Tools used for editing. "editor" allows a full screen editor, "select" provides a cropping and zoom tool.
        type (str): The format the image is converted to before being passed into the prediction function. "numpy" converts the image to a numpy array with shape (width, height, 3) and values from 0 to 255, "pil" converts the image to a PIL image object, "file" produces a temporary file object whose path can be retrieved by file_obj.name, "filepath" returns the path directly.
        label (str): component name in interface.
        """
        self.type = type
        self.default_value = (
            processing_utils.encode_url_or_file_to_base64(default_value)
            if default_value
            else None
        )
        self.type = type
        self.output_type = "auto"
        self.shape = shape
        self.image_mode = image_mode
        self.source = source
        requires_permissions = source == "webcam"
        self.tool = tool
        self.invert_colors = invert_colors
        self.test_input = deepcopy(media_data.BASE64_IMAGE)
        self.interpret_by_tokens = True
        IOComponent.__init__(
            self,
            label=label,
            css=css,
            requires_permissions=requires_permissions,
            **kwargs,
        )

    def get_template_context(self):
        return {
            "image_mode": self.image_mode,
            "shape": self.shape,
            "source": self.source,
            "tool": self.tool,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: Optional[str]) -> np.array | PIL.Image | str | None:
        """
        Parameters:
        x (str): base64 url data
        Returns:
        (Union[numpy.array, PIL.Image, filepath]): image in requested format
        """
        if x is None:
            return x
        im = processing_utils.decode_base64_to_image(x)
        fmt = im.format
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            im = im.convert(self.image_mode)
        if self.shape is not None:
            im = processing_utils.resize_and_crop(im, self.shape)
        if self.invert_colors:
            im = PIL.ImageOps.invert(im)
        if self.type == "pil":
            return im
        elif self.type == "numpy":
            return np.array(im)
        elif self.type == "file" or self.type == "filepath":
            file_obj = tempfile.NamedTemporaryFile(
                delete=False,
                suffix=("." + fmt.lower() if fmt is not None else ".png"),
            )
            im.save(file_obj.name)
            if self.type == "file":
                warnings.warn(
                    "The 'file' type has been deprecated. Set parameter 'type' to 'filepath' instead.",
                    DeprecationWarning,
                )
                return file_obj
            else:
                return file_obj.name
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'numpy', 'pil', 'filepath'."
            )

    def preprocess_example(self, x):
        return processing_utils.encode_file_to_base64(x)

    def serialize(self, x, called_directly=False):
        # if called directly, can assume it's a URL or filepath
        if self.type == "filepath" or called_directly:
            return processing_utils.encode_url_or_file_to_base64(x)
        elif self.type == "file":
            return processing_utils.encode_url_or_file_to_base64(x.name)
        elif self.type in ("numpy", "pil"):
            if self.type == "numpy":
                x = PIL.Image.fromarray(np.uint8(x)).convert("RGB")
            fmt = x.format
            file_obj = tempfile.NamedTemporaryFile(
                delete=False,
                suffix=("." + fmt.lower() if fmt is not None else ".png"),
            )
            x.save(file_obj.name)
            return processing_utils.encode_url_or_file_to_base64(file_obj.name)
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'numpy', 'pil', 'filepath'."
            )

    def set_interpret_parameters(self, segments=16):
        """
        Calculates interpretation score of image subsections by splitting the image into subsections, then using a "leave one out" method to calculate the score of each subsection by whiting out the subsection and measuring the delta of the output value.
        Parameters:
        segments (int): Number of interpretation segments to split image into.
        """
        self.interpretation_segments = segments
        return self

    def _segment_by_slic(self, x):
        """
        Helper method that segments an image into superpixels using slic.
        Parameters:
        x: base64 representation of an image
        """
        x = processing_utils.decode_base64_to_image(x)
        if self.shape is not None:
            x = processing_utils.resize_and_crop(x, self.shape)
        resized_and_cropped_image = np.array(x)
        try:
            from skimage.segmentation import slic
        except (ImportError, ModuleNotFoundError):
            raise ValueError(
                "Error: running this interpretation for images requires scikit-image, please install it first."
            )
        try:
            segments_slic = slic(
                resized_and_cropped_image,
                self.interpretation_segments,
                compactness=10,
                sigma=1,
                start_label=1,
            )
        except TypeError:  # For skimage 0.16 and older
            segments_slic = slic(
                resized_and_cropped_image,
                self.interpretation_segments,
                compactness=10,
                sigma=1,
            )
        return segments_slic, resized_and_cropped_image

    def tokenize(self, x):
        """
        Segments image into tokens, masks, and leave-one-out-tokens
        Parameters:
        x: base64 representation of an image
        Returns:
        tokens: list of tokens, used by the get_masked_input() method
        leave_one_out_tokens: list of left-out tokens, used by the get_interpretation_neighbors() method
        masks: list of masks, used by the get_interpretation_neighbors() method
        """
        segments_slic, resized_and_cropped_image = self._segment_by_slic(x)
        tokens, masks, leave_one_out_tokens = [], [], []
        replace_color = np.mean(resized_and_cropped_image, axis=(0, 1))
        for (i, segment_value) in enumerate(np.unique(segments_slic)):
            mask = segments_slic == segment_value
            image_screen = np.copy(resized_and_cropped_image)
            image_screen[segments_slic == segment_value] = replace_color
            leave_one_out_tokens.append(
                processing_utils.encode_array_to_base64(image_screen)
            )
            token = np.copy(resized_and_cropped_image)
            token[segments_slic != segment_value] = 0
            tokens.append(token)
            masks.append(mask)
        return tokens, leave_one_out_tokens, masks

    def get_masked_inputs(self, tokens, binary_mask_matrix):
        masked_inputs = []
        for binary_mask_vector in binary_mask_matrix:
            masked_input = np.zeros_like(tokens[0], dtype=int)
            for token, b in zip(tokens, binary_mask_vector):
                masked_input = masked_input + token * int(b)
            masked_inputs.append(processing_utils.encode_array_to_base64(masked_input))
        return masked_inputs

    def get_interpretation_scores(
        self, x, neighbors, scores, masks, tokens=None, **kwargs
    ):
        """
        Returns:
        (List[List[float]]): A 2D array representing the interpretation score of each pixel of the image.
        """
        x = processing_utils.decode_base64_to_image(x)
        if self.shape is not None:
            x = processing_utils.resize_and_crop(x, self.shape)
        x = np.array(x)
        output_scores = np.zeros((x.shape[0], x.shape[1]))

        for score, mask in zip(scores, masks):
            output_scores += score * mask

        max_val, min_val = np.max(output_scores), np.min(output_scores)
        if max_val > 0:
            output_scores = (output_scores - min_val) / (max_val - min_val)
        return output_scores.tolist()

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (str) path to image file
        """
        return self.save_flagged_file(dir, label, data, encryption_key)

    def restore_flagged(self, dir, data, encryption_key):
        return os.path.join(dir, data)

    def generate_sample(self):
        return deepcopy(media_data.BASE64_IMAGE)

    # Output functions

    def postprocess(self, y):
        """
        Parameters:
        y (Union[numpy.array, PIL.Image, str, matplotlib.pyplot, Tuple[Union[numpy.array, PIL.Image, str], List[Tuple[str, float, float, float, float]]]]): image in specified format
        Returns:
        (str): base64 url data
        """
        if self.output_type == "auto":
            if isinstance(y, np.ndarray):
                dtype = "numpy"
            elif isinstance(y, PIL.Image.Image):
                dtype = "pil"
            elif isinstance(y, str):
                dtype = "file"
            elif isinstance(y, (ModuleType, matplotlib.figure.Figure)):
                dtype = "plot"
            else:
                raise ValueError(
                    "Unknown type. Please choose from: 'numpy', 'pil', 'file', 'plot'."
                )
        else:
            dtype = self.output_type
        if dtype in ["numpy", "pil"]:
            if dtype == "pil":
                y = np.array(y)
            out_y = processing_utils.encode_array_to_base64(y)
        elif dtype == "file":
            out_y = processing_utils.encode_url_or_file_to_base64(y)
        elif dtype == "plot":
            out_y = processing_utils.encode_plot_to_base64(y)
        else:
            raise ValueError(
                "Unknown type: "
                + dtype
                + ". Please choose from: 'numpy', 'pil', 'file', 'plot'."
            )
        return out_y

    def deserialize(self, x):
        y = processing_utils.decode_base64_to_file(x).name
        return y


class Video(Changeable, Clearable, Playable, IOComponent):
    """
    Component creates a video file upload that is converted to a file path.

    Input type: filepath
    Output type: filepath
    Demos: video_flip
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        type: Optional[str] = None,
        source: str = "upload",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value(str): A path or URL for the default value that Video component is going to take.
        type (str): Type of video format to be returned by component, such as 'avi' or 'mp4'. Use 'mp4' to ensure browser playability. If set to None, video will keep uploaded format.
        source (str): Source of video. "upload" creates a box where user can drop an video file, "webcam" allows user to record a video from their webcam.
        label (str): component name in interface.
        optional (bool): If True, the interface can be submitted with no uploaded video, in which case the input value is None.
        """
        self.default_value = (
            processing_utils.encode_url_or_file_to_base64(default_value)
            if default_value
            else None
        )
        self.type = type
        self.source = source
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "source": self.source,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess_example(self, x):
        return {"name": x, "data": None, "is_example": True}

    def preprocess(self, x: Dict[str, str] | None) -> str | None:
        """
        Parameters:
        x (Dict[name: str, data: str]): JSON object with filename as 'name' property and base64 data as 'data' property
        Returns:
        (str): file path to video
        """
        if x is None:
            return x
        file_name, file_data, is_example = (
            x["name"],
            x["data"],
            x.get("is_example", False),
        )
        if is_example:
            file = processing_utils.create_tmp_copy_of_file(file_name)
        else:
            file = processing_utils.decode_base64_to_file(
                file_data, file_path=file_name
            )
        file_name = file.name
        uploaded_format = file_name.split(".")[-1].lower()
        if self.type is not None and uploaded_format != self.type:
            output_file_name = file_name[0 : file_name.rindex(".") + 1] + self.type
            ff = FFmpeg(inputs={file_name: None}, outputs={output_file_name: None})
            ff.run()
            return output_file_name
        else:
            return file_name

    def serialize(self, x, called_directly):
        raise NotImplementedError()

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (str) path to video file
        """
        return self.save_flagged_file(
            dir, label, None if data is None else data["data"], encryption_key
        )

    def generate_sample(self):
        return deepcopy(media_data.BASE64_VIDEO)

    def postprocess(self, y):
        """
        Parameters:
        y (str): path to video
        Returns:
        (str): base64 url data
        """
        returned_format = y.split(".")[-1].lower()
        if self.type is not None and returned_format != self.type:
            output_file_name = y[0 : y.rindex(".") + 1] + self.type
            ff = FFmpeg(inputs={y: None}, outputs={output_file_name: None})
            ff.run()
            y = output_file_name
        return {
            "name": os.path.basename(y),
            "data": processing_utils.encode_file_to_base64(y),
        }

    def deserialize(self, x):
        return processing_utils.decode_base64_to_file(x).name


class Audio(Changeable, Clearable, Playable, IOComponent):
    """
    Component accepts audio input files or creates an audio player that plays the output audio.


    Input type: Union[Tuple[int, numpy.array], file-object, numpy.array]
    Output type: Union[Tuple[int, numpy.array], str]
    Demos: main_note, generate_tone, reverse_audio, spectogram
    """

    def __init__(
        self,
        default_value="",
        *,
        source: str = "upload",
        type: str = "numpy",
        label: str = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): IGNORED
        source (str): Source of audio. "upload" creates a box where user can drop an audio file, "microphone" creates a microphone input.
        type (str): The format the image is converted to before being passed into the prediction function. "numpy" converts the image to a numpy array with shape (width, height, 3) and values from 0 to 255, "pil" converts the image to a PIL image object, "file" produces a temporary file object whose path can be retrieved by file_obj.name, "filepath" returns the path directly.
        label (str): component name in interface.
        """
        self.default_value = (
            processing_utils.encode_url_or_file_to_base64(default_value)
            if default_value
            else None
        )
        self.source = source
        requires_permissions = source == "microphone"
        self.type = type
        self.output_type = "auto"
        self.test_input = deepcopy(media_data.BASE64_AUDIO)
        self.interpret_by_tokens = True
        IOComponent.__init__(
            self, label=label, requires_permissions=requires_permissions, **kwargs
        )

    def get_template_context(self):
        return {
            "source": self.source,  # TODO: This did not exist in output template, careful here if an error arrives
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess_example(self, x):
        return {"name": x, "data": None, "is_example": True}

    def preprocess(self, x: Dict[str, str] | None) -> Tuple[int, np.array] | str | None:
        """
        Parameters:
        x (Dict[name: str, data: str]): JSON object with filename as 'name' property and base64 data as 'data' property
        Returns:
        (Union[Tuple[int, numpy.array], str, numpy.array]): audio in requested format
        """
        if x is None:
            return x
        file_name, file_data, is_example = (
            x["name"],
            x["data"],
            x.get("is_example", False),
        )
        crop_min, crop_max = x.get("crop_min", 0), x.get("crop_max", 100)
        if is_example:
            file_obj = processing_utils.create_tmp_copy_of_file(file_name)
        else:
            file_obj = processing_utils.decode_base64_to_file(
                file_data, file_path=file_name
            )
        if crop_min != 0 or crop_max != 100:
            sample_rate, data = processing_utils.audio_from_file(
                file_obj.name, crop_min=crop_min, crop_max=crop_max
            )
            processing_utils.audio_to_file(sample_rate, data, file_obj.name)
        if self.type == "file":
            warnings.warn(
                "The 'file' type has been deprecated. Set parameter 'type' to 'filepath' instead.",
                DeprecationWarning,
            )
            return file_obj
        elif self.type == "filepath":
            return file_obj.name
        elif self.type == "numpy":
            return processing_utils.audio_from_file(file_obj.name)
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'numpy', 'filepath'."
            )

    def serialize(self, x, called_directly):
        if x is None:
            return None
        if self.type == "filepath" or called_directly:
            name = x
        elif self.type == "file":
            warnings.warn(
                "The 'file' type has been deprecated. Set parameter 'type' to 'filepath' instead.",
                DeprecationWarning,
            )
            name = x.name
        elif self.type == "numpy":
            file = tempfile.NamedTemporaryFile(delete=False)
            name = file.name
            processing_utils.audio_to_file(x[0], x[1], name)
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'numpy', 'filepath'."
            )

        file_data = processing_utils.encode_url_or_file_to_base64(name)
        return {"name": name, "data": file_data, "is_example": False}

    def set_interpret_parameters(self, segments=8):
        """
        Calculates interpretation score of audio subsections by splitting the audio into subsections, then using a "leave one out" method to calculate the score of each subsection by removing the subsection and measuring the delta of the output value.
        Parameters:
        segments (int): Number of interpretation segments to split audio into.
        """
        self.interpretation_segments = segments
        return self

    def tokenize(self, x):
        if x.get("is_example"):
            sample_rate, data = processing_utils.audio_from_file(x["name"])
        else:
            file_obj = processing_utils.decode_base64_to_file(x["data"])
            sample_rate, data = processing_utils.audio_from_file(file_obj.name)
        leave_one_out_sets = []
        tokens = []
        masks = []
        duration = data.shape[0]
        boundaries = np.linspace(0, duration, self.interpretation_segments + 1).tolist()
        boundaries = [round(boundary) for boundary in boundaries]
        for index in range(len(boundaries) - 1):
            start, stop = boundaries[index], boundaries[index + 1]
            masks.append((start, stop))

            # Handle the leave one outs
            leave_one_out_data = np.copy(data)
            leave_one_out_data[start:stop] = 0
            file = tempfile.NamedTemporaryFile(delete=False, suffix=".wav")
            processing_utils.audio_to_file(sample_rate, leave_one_out_data, file.name)
            out_data = processing_utils.encode_file_to_base64(file.name)
            leave_one_out_sets.append(out_data)
            file.close()
            os.unlink(file.name)

            # Handle the tokens
            token = np.copy(data)
            token[0:start] = 0
            token[stop:] = 0
            file = tempfile.NamedTemporaryFile(delete=False, suffix=".wav")
            processing_utils.audio_to_file(sample_rate, token, file.name)
            token_data = processing_utils.encode_file_to_base64(file.name)
            file.close()
            os.unlink(file.name)

            tokens.append(token_data)
        tokens = [{"name": "token.wav", "data": token} for token in tokens]
        leave_one_out_sets = [
            {"name": "loo.wav", "data": loo_set} for loo_set in leave_one_out_sets
        ]
        return tokens, leave_one_out_sets, masks

    def get_masked_inputs(self, tokens, binary_mask_matrix):
        # create a "zero input" vector and get sample rate
        x = tokens[0]["data"]
        file_obj = processing_utils.decode_base64_to_file(x)
        sample_rate, data = processing_utils.audio_from_file(file_obj.name)
        zero_input = np.zeros_like(data, dtype="int16")
        # decode all of the tokens
        token_data = []
        for token in tokens:
            file_obj = processing_utils.decode_base64_to_file(token["data"])
            _, data = processing_utils.audio_from_file(file_obj.name)
            token_data.append(data)
        # construct the masked version
        masked_inputs = []
        for binary_mask_vector in binary_mask_matrix:
            masked_input = np.copy(zero_input)
            for t, b in zip(token_data, binary_mask_vector):
                masked_input = masked_input + t * int(b)
            file = tempfile.NamedTemporaryFile(delete=False)
            processing_utils.audio_to_file(sample_rate, masked_input, file.name)
            masked_data = processing_utils.encode_file_to_base64(file.name)
            file.close()
            os.unlink(file.name)
            masked_inputs.append(masked_data)
        return masked_inputs

    def get_interpretation_scores(self, x, neighbors, scores, masks=None, tokens=None):
        """
        Returns:
        (List[float]): Each value represents the interpretation score corresponding to an evenly spaced subsection of audio.
        """
        return list(scores)

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (str) path to audio file
        """
        if data is None:
            data_string = None
        elif isinstance(data, str):
            data_string = data
        else:
            data_string = data["data"]
            is_example = data.get("is_example", False)
            if is_example:
                file_obj = processing_utils.create_tmp_copy_of_file(data["name"])
                return self.save_file(file_obj, dir, label)

        return self.save_flagged_file(dir, label, data_string, encryption_key)

    def generate_sample(self):
        return deepcopy(media_data.BASE64_AUDIO)

    def postprocess(self, y):
        """
        Parameters:
        y (Union[Tuple[int, numpy.array], str]): audio data in requested format
        Returns:
        (str): base64 url data
        """
        if self.output_type in ["numpy", "file", "auto"]:
            if self.output_type == "numpy" or (
                self.output_type == "auto" and isinstance(y, tuple)
            ):
                sample_rate, data = y
                file = tempfile.NamedTemporaryFile(
                    prefix="sample", suffix=".wav", delete=False
                )
                processing_utils.audio_to_file(sample_rate, data, file.name)
                y = file.name
            return processing_utils.encode_url_or_file_to_base64(y)
        else:
            raise ValueError(
                "Unknown type: " + self.type + ". Please choose from: 'numpy', 'file'."
            )

    def deserialize(self, x):
        return processing_utils.decode_base64_to_file(x).name


class File(Changeable, Clearable, IOComponent):
    """
    Component accepts generic file uploads and output..

    Input type: Union[file-object, bytes, List[Union[file-object, bytes]]]
    Output type: Union[file-like, str]
    Demos: zip_to_json, zip_two_files
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        file_count: str = "single",
        type: str = "file",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value given as file path
        file_count (str): if single, allows user to upload one file. If "multiple", user uploads multiple files. If "directory", user uploads all files in selected directory. Return type will be list for each file in case of "multiple" or "directory".
        type (str): Type of value to be returned by component. "file" returns a temporary file object whose path can be retrieved by file_obj.name, "binary" returns an bytes object.
        label (str): component name in interface.
        """
        self.default_value = (
            processing_utils.encode_url_or_file_to_base64(default_value)
            if default_value
            else None
        )
        self.file_count = file_count
        self.type = type
        self.test_input = None
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "file_count": self.file_count,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def preprocess_example(self, x):
        return {"name": x, "data": None, "is_example": True}

    def preprocess(self, x: List[Dict[str, str]] | None):
        """
        Parameters:
        x (List[Dict[name: str, data: str]]): List of JSON objects with filename as 'name' property and base64 data as 'data' property
        Returns:
        (Union[file-object, bytes, List[Union[file-object, bytes]]]): File objects in requested format
        """
        if x is None:
            return None

        def process_single_file(f):
            file_name, data, is_example = (
                f["name"],
                f["data"],
                f.get("is_example", False),
            )
            if self.type == "file":
                if is_example:
                    return processing_utils.create_tmp_copy_of_file(file_name)
                else:
                    return processing_utils.decode_base64_to_file(
                        data, file_path=file_name
                    )
            elif self.type == "bytes":
                if is_example:
                    with open(file_name, "rb") as file_data:
                        return file_data.read()
                return processing_utils.decode_base64_to_binary(data)[0]
            else:
                raise ValueError(
                    "Unknown type: "
                    + str(self.type)
                    + ". Please choose from: 'file', 'bytes'."
                )

        if self.file_count == "single":
            if isinstance(x, list):
                return process_single_file(x[0])
            else:
                return process_single_file(x)
        else:
            return [process_single_file(f) for f in x]

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (str) path to file
        """
        return self.save_flagged_file(
            dir, label, None if data is None else data[0]["data"], encryption_key
        )

    def generate_sample(self):
        return deepcopy(media_data.BASE64_FILE)

    # Output Functionalities

    def postprocess(self, y):
        """
        Parameters:
        y (str): file path
        Returns:
        (Dict[name: str, size: number, data: str]): JSON object with key 'name' for filename, 'data' for base64 url, and 'size' for filesize in bytes
        """
        return {
            "name": os.path.basename(y),
            "size": os.path.getsize(y),
            "data": processing_utils.encode_file_to_base64(y),
        }


class Dataframe(Changeable, IOComponent):
    """
    Component accepts or displays 2D input  through a spreadsheet interface.

    Input or Output type: Union[pandas.DataFrame, numpy.array, List[Union[str, float]], List[List[Union[str, float]]]]
    Demos: filter_records, matrix_transpose, tax_calculator
    """

    def __init__(
        self,
        default_value: Optional[List[List[Any]]] = None,
        *,
        headers: Optional[List[str]] = None,
        row_count: int = 3,
        col_count: Optional[int] = 3,
        datatype: str | List[str] = "str",
        col_width: int | List[int] = None,
        type: str = "pandas",
        label: Optional[str] = None,
        max_rows: Optional[int] = 20,
        max_cols: Optional[int] = None,
        overflow_row_behaviour: str = "paginate",
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Input Parameters:
        default_value (List[List[Any]]): Default value as a pandas DataFrame. TODO: Add support for default value as a filepath
        headers (List[str]): Header names to dataframe. If None, no headers are shown.
        row_count (int): Limit number of rows for input.
        col_count (int): Limit number of columns for input. If equal to 1, return data will be one-dimensional. Ignored if `headers` is provided.
        datatype (Union[str, List[str]]): Datatype of values in sheet. Can be provided per column as a list of strings, or for the entire sheet as a single string. Valid datatypes are "str", "number", "bool", and "date".
        col_width (Union[int, List[int]]): Width of columns in pixels. Can be provided as single value or list of values per column.
        type (str): Type of value to be returned by component. "pandas" for pandas dataframe, "numpy" for numpy array, or "array" for a Python array.
        label (str): component name in interface.
        Output Parameters:
        headers (List[str]): Header names to dataframe. Only applicable if type is "numpy" or "array".
        max_rows (int): Maximum number of rows to display at once. Set to None for infinite.
        max_cols (int): Maximum number of columns to display at once. Set to None for infinite.
        overflow_row_behaviour (str): If set to "paginate", will create pages for overflow rows. If set to "show_ends", will show initial and final rows and truncate middle rows.
        """
        self.headers = headers
        self.datatype = datatype
        self.row_count = row_count
        self.col_count = len(headers) if headers else col_count
        self.col_width = col_width
        self.type = type
        self.output_type = "auto"
        default_values = {
            "str": "",
            "number": 0,
            "bool": False,
            "date": "01/01/1970",
        }
        column_dtypes = (
            [datatype] * self.col_count if isinstance(datatype, str) else datatype
        )
        self.test_input = [
            [default_values[c] for c in column_dtypes] for _ in range(row_count)
        ]
        self.default_value = (
            default_value if default_value is not None else self.test_input
        )
        self.max_rows = max_rows
        self.max_cols = max_cols
        self.overflow_row_behaviour = overflow_row_behaviour
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "headers": self.headers,
            "datatype": self.datatype,
            "row_count": self.row_count,
            "col_count": self.col_count,
            "col_width": self.col_width,
            "default_value": self.default_value,
            "max_rows": self.max_rows,
            "max_cols": self.max_cols,
            "overflow_row_behaviour": self.overflow_row_behaviour,
            **IOComponent.get_template_context(self),
        }

    def preprocess(self, x: List[List[str | Number | bool]]):
        """
        Parameters:
        x (List[List[Union[str, number, bool]]]): 2D array of str, numeric, or bool data
        Returns:
        (Union[pandas.DataFrame, numpy.array, List[Union[str, float]], List[List[Union[str, float]]]]): Dataframe in requested format
        """
        if self.type == "pandas":
            if self.headers:
                return pd.DataFrame(x, columns=self.headers)
            else:
                return pd.DataFrame(x)
        if self.col_count == 1:
            x = [row[0] for row in x]
        if self.type == "numpy":
            return np.array(x)
        elif self.type == "array":
            return x
        else:
            raise ValueError(
                "Unknown type: "
                + str(self.type)
                + ". Please choose from: 'pandas', 'numpy', 'array'."
            )

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (List[List[Union[str, float]]]) 2D array
        """
        return json.dumps(data)
        # TODO: (faruk) output was dumping differently, how to converge?
        # return json.dumps(data["data"])

    def restore_flagged(self, dir, data, encryption_key):
        return json.loads(data)
        # TODO: (faruk) output was dumping differently, how to converge?
        # return {"data": json.loads(data)}

    def generate_sample(self):
        return [[1, 2, 3], [4, 5, 6]]

    def postprocess(self, y):
        """
        Parameters:
        y (Union[pandas.DataFrame, numpy.array, List[Union[str, float]], List[List[Union[str, float]]]]): dataframe in given format
        Returns:
        (Dict[headers: List[str], data: List[List[Union[str, number]]]]): JSON object with key 'headers' for list of header names, 'data' for 2D array of string or numeric data
        """
        if self.output_type == "auto":
            if isinstance(y, pd.core.frame.DataFrame):
                dtype = "pandas"
            elif isinstance(y, np.ndarray):
                dtype = "numpy"
            elif isinstance(y, list):
                dtype = "array"
            else:
                raise ValueError("Cannot determine the type of DataFrame output.")
        else:
            dtype = self.output_type
        if dtype == "pandas":
            return {"headers": list(y.columns), "data": y.values.tolist()}
        elif dtype in ("numpy", "array"):
            if dtype == "numpy":
                y = y.tolist()
            if len(y) == 0 or not isinstance(y[0], list):
                y = [y]
            return {"data": y}
        else:
            raise ValueError(
                "Unknown type: "
                + self.type
                + ". Please choose from: 'pandas', 'numpy', 'array'."
            )


class Timeseries(Changeable, IOComponent):
    """
    Component accepts pandas.DataFrame uploaded as a timeseries csv file or renders a dataframe consisting of a time series as output.

    Input type: pandas.DataFrame
    Output type: pandas.DataFrame
    Demos: fraud_detector
    """

    def __init__(
        self,
        default_value: Optional[str] = None,
        *,
        x: Optional[str] = None,
        y: str | List[str] = None,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        colors: List[str] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value: File path for the timeseries csv file. TODO: Add support for default value as a pd.DataFrame
        x (str): Column name of x (time) series. None if csv has no headers, in which case first column is x series.
        y (Union[str, List[str]]): Column name of y series, or list of column names if multiple series. None if csv has no headers, in which case every column after first is a y series.
        label (str): component name in interface.
        colors List[str]: an ordered list of colors to use for each line plot
        """
        self.default_value = (
            pd.read_csv(default_value) if default_value is not None else None
        )
        self.x = x
        if isinstance(y, str):
            y = [y]
        self.y = y
        self.colors = colors
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "x": self.x,
            "y": self.y,
            "default_value": self.default_value,
            "colors": self.colors,
            **IOComponent.get_template_context(self),
        }

    def preprocess_example(self, x):
        return {"name": x, "is_example": True}

    def preprocess(self, x: Dict | None) -> pd.DataFrame | None:
        """
        Parameters:
        x (Dict[data: List[List[Union[str, number, bool]]], headers: List[str], range: List[number]]): Dict with keys 'data': 2D array of str, numeric, or bool data, 'headers': list of strings for header names, 'range': optional two element list designating start of end of subrange.
        Returns:
        (pandas.DataFrame): Dataframe of timeseries data
        """
        if x is None:
            return x
        elif x.get("is_example"):
            dataframe = pd.read_csv(x["name"])
        else:
            dataframe = pd.DataFrame(data=x["data"], columns=x["headers"])
        if x.get("range") is not None:
            dataframe = dataframe.loc[dataframe[self.x or 0] >= x["range"][0]]
            dataframe = dataframe.loc[dataframe[self.x or 0] <= x["range"][1]]
        return dataframe

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (List[List[Union[str, float]]]) 2D array
        """
        return json.dumps(data)

    def restore_flagged(self, dir, data, encryption_key):
        return json.loads(data)

    def generate_sample(self):
        return {"data": [[1] + [2] * len(self.y)] * 4, "headers": [self.x] + self.y}

    # Output Functionalities

    def postprocess(self, y):
        """
        Parameters:
        y (pandas.DataFrame): timeseries data
        Returns:
        (Dict[headers: List[str], data: List[List[Union[str, number]]]]): JSON object with key 'headers' for list of header names, 'data' for 2D array of string or numeric data
        """
        return {"headers": y.columns.values.tolist(), "data": y.values.tolist()}


class Variable(IOComponent):
    """
    Special hidden component that stores state across runs of the interface.

    Input type: Any
    Output type: Any
    Demos: chatbot
    """

    def __init__(
        self,
        default_value: Any = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (Any): the initial value of the state.
        label (str): component name in interface (not used).
        """
        self.default_value = default_value
        self.stateful = True
        IOComponent.__init__(self, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }


############################
# Only Output Components
############################


class Label(Changeable, IOComponent):
    """
    Component outputs a classification label, along with confidence scores of top categories if provided. Confidence scores are represented as a dictionary mapping labels to scores between 0 and 1.
    Output type: Union[Dict[str, float], str, int, float]
    Demos: image_classifier, main_note, titanic_survival
    """

    CONFIDENCES_KEY = "confidences"

    def __init__(
        self,
        default_value: str = "",
        *,
        num_top_classes: Optional[int] = None,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value(str): Default string value
        num_top_classes (int): number of most confident classes to show.
        label (str): component name in interface.
        """
        # TODO: Shall we have a default value for the label component?
        self.num_top_classes = num_top_classes
        self.output_type = "auto"
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def postprocess(self, y):
        """
        Parameters:
        y (Dict[str, float]): dictionary mapping label to confidence value
        Returns:
        (Dict[label: str, confidences: List[Dict[label: str, confidence: number]]]): Object with key 'label' representing primary label, and key 'confidences' representing a list of label-confidence pairs
        """
        if self.output_type == "label" or (
            self.output_type == "auto" and (isinstance(y, (str, numbers.Number)))
        ):
            return {"label": str(y)}
        elif self.output_type == "confidences" or (
            self.output_type == "auto" and isinstance(y, dict)
        ):
            sorted_pred = sorted(y.items(), key=operator.itemgetter(1), reverse=True)
            if self.num_top_classes is not None:
                sorted_pred = sorted_pred[: self.num_top_classes]
            return {
                "label": sorted_pred[0][0],
                "confidences": [
                    {"label": pred[0], "confidence": pred[1]} for pred in sorted_pred
                ],
            }
        else:
            raise ValueError(
                "The `Label` output interface expects one of: a string label, or an int label, a "
                "float label, or a dictionary whose keys are labels and values are confidences. "
                "Instead, got a {}".format(type(y))
            )

    def deserialize(self, y):
        # 5 cases: (1): {'label': 'lion'}, {'label': 'lion', 'confidences':...}, {'lion': 0.46, ...}, 'lion', '0.46'
        if self.output_type == "label" or (
            self.output_type == "auto"
            and (
                isinstance(y, (str, numbers.Number))
                or ("label" in y and not ("confidences" in y.keys()))
            )
        ):
            if isinstance(y, (str, numbers.Number)):
                return y
            else:
                return y["label"]
        elif self.output_type == "confidences" or self.output_type == "auto":
            if ("confidences" in y.keys()) and isinstance(y["confidences"], list):
                return {k["label"]: k["confidence"] for k in y["confidences"]}
            else:
                return y
        raise ValueError("Unable to deserialize output: {}".format(y))

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (Union[str, Dict[str, number]]): Either a string representing the main category label, or a dictionary with category keys mapping to confidence levels.
        """
        if "confidences" in data:
            return json.dumps(
                {
                    example["label"]: example["confidence"]
                    for example in data["confidences"]
                }
            )
        else:
            return data["label"]

    def restore_flagged(self, dir, data, encryption_key):
        try:
            data = json.loads(data)
            return self.postprocess(data)
        except ValueError:
            return data


class KeyValues(IOComponent):
    """
    Component displays a table representing values for multiple fields.
    Output type: Union[Dict, List[Tuple[str, Union[str, int, float]]]]
    Demos: text_analysis
    """

    def __init__(
        self,
        default_value: str = " ",
        *,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default (str): IGNORED
        label (str): component name in interface.
        """
        raise DeprecationWarning(
            "The KeyValues component is deprecated. Please use the DataFrame or JSON "
            "components instead."
        )


class HighlightedText(Changeable, IOComponent):
    """
    Component creates text that contains spans that are highlighted by category or numerical value.
    Output is represent as a list of Tuple pairs, where the first element represents the span of text represented by the tuple, and the second element represents the category or value of the text.
    Output type: List[Tuple[str, Union[float, str]]]
    Demos: diff_texts, text_analysis
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        color_map: Dict[str, str] = None,
        label: Optional[str] = None,
        show_legend: bool = False,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        color_map (Dict[str, str]): Map between category and respective colors
        label (str): component name in interface.
        show_legend (bool): whether to show span categories in a separate legend or inline.
        """
        self.default_value = default_value
        self.color_map = color_map
        self.show_legend = show_legend
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "color_map": self.color_map,
            "show_legend": self.show_legend,
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def postprocess(self, y):
        """
        Parameters:
        y (Union[Dict, List[Tuple[str, Union[str, int, float]]]]): dictionary or tuple list representing key value pairs
        Returns:
        (List[Tuple[str, Union[str, number]]]): list of key value pairs

        """
        return y

    def save_flagged(self, dir, label, data, encryption_key):
        return json.dumps(data)

    def restore_flagged(self, dir, data, encryption_key):
        return json.loads(data)


class JSON(Changeable, IOComponent):
    """
    Used for JSON output. Expects a JSON string or a Python object that is JSON serializable.
    Output type: Union[str, Any]
    Demos: zip_to_json
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        label (str): component name in interface.
        """
        self.default_value = json.dumps(default_value)
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def postprocess(self, y):
        """
        Parameters:
        y (Union[Dict, List, str]): JSON output
        Returns:
        (Union[Dict, List]): JSON output
        """
        if isinstance(y, str):
            return json.dumps(y)
        else:
            return y

    def save_flagged(self, dir, label, data, encryption_key):
        return json.dumps(data)

    def restore_flagged(self, dir, data, encryption_key):
        return json.loads(data)


class HTML(Changeable, IOComponent):
    """
    Used for HTML output. Expects an HTML valid string.
    Output type: str
    Demos: text_analysis
    """

    def __init__(
        self,
        default_value: str = "",
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        label (str): component name in interface.
        """
        self.default_value = default_value
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **IOComponent.get_template_context(self),
        }

    def postprocess(self, x):
        """
        Parameters:
        y (str): HTML output
        Returns:
        (str): HTML output
        """
        return x


class Gallery(IOComponent):
    def __init__(
        self,
        *,
        label: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(label=label, **kwargs)

    def get_template_context(self):
        return {
            **super().get_template_context(),
        }

    def postprocess(self, y):
        """
        Parameters:
        y (List[Union[numpy.array, PIL.Image, str]]): list of images
        Returns:
        (str): list of base64 url data for images
        """
        output = []
        for img in y:
            if isinstance(img, np.ndarray):
                img = processing_utils.encode_array_to_base64(img)
            elif isinstance(img, PIL.Image.Image):
                img = np.array(img)
                img = processing_utils.encode_array_to_base64(img)
            elif isinstance(img, str):
                img = processing_utils.encode_url_or_file_to_base64(img)
            else:
                raise ValueError(
                    "Unknown type. Please choose from: 'numpy', 'pil', 'file'."
                )
            output.append(img)
        return output


class Carousel(IOComponent):
    """
    Component displays a set of output components that can be scrolled through.
    Output type: List[List[Any]]
    Demos: disease_report
    """

    def __init__(
        self,
        *,
        components: Component | List[Component],
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        components (Union[List[OutputComponent], OutputComponent]): Classes of component(s) that will be scrolled through.
        label (str): component name in interface.
        """
        if not isinstance(components, list):
            components = [components]
        self.components = [
            get_component_instance(component) for component in components
        ]
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "components": [
                component.get_template_context() for component in self.components
            ],
            **IOComponent.get_template_context(self),
        }

    def postprocess(self, y):
        """
        Parameters:
        y (List[List[Any]]): carousel output
        Returns:
        (List[List[Any]]): 2D array, where each sublist represents one set of outputs or 'slide' in the carousel
        """
        if isinstance(y, list):
            if len(y) != 0 and not isinstance(y[0], list):
                y = [[z] for z in y]
            output = []
            for row in y:
                output_row = []
                for i, cell in enumerate(row):
                    output_row.append(self.components[i].postprocess(cell))
                output.append(output_row)
            return output
        else:
            raise ValueError("Unknown type. Please provide a list for the Carousel.")

    def save_flagged(self, dir, label, data, encryption_key):
        return json.dumps(
            [
                [
                    component.save_flagged(
                        dir, f"{label}_{j}", data[i][j], encryption_key
                    )
                    for j, component in enumerate(self.components)
                ]
                for i, _ in enumerate(data)
            ]
        )

    def restore_flagged(self, dir, data, encryption_key):
        return [
            [
                component.restore_flagged(dir, sample, encryption_key)
                for component, sample in zip(self.components, sample_set)
            ]
            for sample_set in json.loads(data)
        ]


class Chatbot(Changeable, IOComponent):
    """
    Component displays a chatbot output showing both user submitted messages and responses
    Output type: List[Tuple[str, str]]
    Demos: chatbot
    """

    def __init__(
        self,
        default_value="",
        color_map: Tuple(str, str) = None,
        *,
        label: Optional[str] = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        color_map (Tuple[str, str]): Chat bubble color of input text and output text respectively.
        label (str): component name in interface (not used).
        """
        self.default_value = default_value
        self.color_map = color_map
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            "color_map": self.color_map,
            **IOComponent.get_template_context(self),
        }

    def postprocess(self, y):
        """
        Parameters:
        y (List[Tuple[str, str]]): List of tuples representing the message and response
        Returns:
        (List[Tuple[str, str]]): Returns same list of tuples

        """
        return y


class Model3D(Changeable, Editable, Clearable, IOComponent):
    """
    Component creates a 3D Model component with input and output capabilities.
    Input type: File object of type (.obj, glb, or .gltf)
    Output type: filepath
    Demos: Model3D
    """

    def __init__(
        self,
        clear_color=None,
        label: str = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        clear_color (List[r, g, b, a]): background color of scene
        label (str): component name in interface.
        """
        self.clear_color = clear_color
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {
            "clearColor": self.clear_color,
            **IOComponent.get_template_context(self),
        }

    def preprocess_example(self, x):
        return {"name": x, "data": None, "is_example": True}

    def preprocess(self, x: Dict[str, str] | None) -> str | None:
        """
        Parameters:
        x (Dict[name: str, data: str]): JSON object with filename as 'name' property and base64 data as 'data' property
        Returns:
        (str): file path to 3D image model
        """
        if x is None:
            return x
        file_name, file_data, is_example = (
            x["name"],
            x["data"],
            x.get("is_example", False),
        )
        if is_example:
            file = processing_utils.create_tmp_copy_of_file(file_name)
        else:
            file = processing_utils.decode_base64_to_file(
                file_data, file_path=file_name
            )
        file_name = file.name
        return file_name

    def serialize(self, x, called_directly):
        raise NotImplementedError()

    def save_flagged(self, dir, label, data, encryption_key):
        """
        Returns: (str) path to 3D image model file
        """
        return self.save_flagged_file(
            dir, label, data["data"], encryption_key, data["name"]
        )

    def generate_sample(self):
        return media_data.BASE64_MODEL3D

    # Output functions

    def postprocess(self, y):
        """
        Parameters:
        y (str): path to the model
        Returns:
        (str): file name
        (str): file extension
        (str): base64 url data
        """

        if self.clear_color is None:
            self.clear_color = [0.2, 0.2, 0.2, 1.0]

        return {
            "name": os.path.basename(y),
            "data": processing_utils.encode_file_to_base64(y),
        }

    def deserialize(self, x):
        return processing_utils.decode_base64_to_file(x).name

    def restore_flagged(self, dir, data, encryption_key):
        return self.restore_flagged_file(dir, data, encryption_key)


class Plot(Changeable, Clearable, IOComponent):
    """
    Used for plot output.
    Output type: matplotlib plt, plotly figure, or Bokeh fig (json_item format)
    Demos: outbreak_forecast
    """

    def __init__(
        self,
        type: str = None,
        label: str = None,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        type (str): type of plot (matplotlib, plotly)
        label (str): component name in interface.
        """
        self.type = type
        IOComponent.__init__(self, label=label, css=css, **kwargs)

    def get_template_context(self):
        return {**IOComponent.get_template_context(self)}

    def postprocess(self, y):
        """
        Parameters:
        y (str): plot data
        Returns:
        (str): plot type
        (str): plot base64 or json
        """
        dtype = self.type
        if self.type == "plotly":
            out_y = y.to_json()
        elif self.type == "matplotlib":
            out_y = processing_utils.encode_plot_to_base64(y)
        elif self.type == "bokeh":
            out_y = json.dumps(y)
        elif self.type == "auto":
            if isinstance(y, (ModuleType, matplotlib.pyplot.Figure)):
                dtype = "matplotlib"
                out_y = processing_utils.encode_plot_to_base64(y)
            elif isinstance(y, dict):
                dtype = "bokeh"
                out_y = json.dumps(y)
            else:
                dtype = "plotly"
                out_y = y.to_json()
        else:
            raise ValueError(
                "Unknown type. Please choose from: 'plotly', 'matplotlib', 'bokeh'."
            )
        return {"type": dtype, "plot": out_y}


############################
# Static Components
############################


class Markdown(Component):
    """
    Used for Markdown output. Expects a valid string that is rendered into Markdown.
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        css (dict): optional css parameters for the component
        """
        Component.__init__(self, css=css, **kwargs)
        self.md = MarkdownIt()
        unindented_default_value = inspect.cleandoc(default_value)
        self.default_value = self.md.render(unindented_default_value)

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **Component.get_template_context(self),
        }


class Button(Clickable, Component):
    """
    Used to create a button, that can be assigned arbitrary click() events.
    """

    def __init__(
        self,
        default_value: str = "",
        *,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        default_value (str): Default value
        css (dict): optional css parameters for the component
        """
        Component.__init__(self, css=css, **kwargs)
        self.default_value = default_value

    def get_template_context(self):
        return {
            "default_value": self.default_value,
            **Component.get_template_context(self),
        }


class Dataset(Clickable, Component):
    """
    Used to create a output widget for showing datasets. Used to render the examples
    box in the interface.
    """

    def __init__(
        self,
        *,
        components: List[Component],
        samples: List[List[Any]],
        type: str = "values",
        css: Optional[CSS] = None,
        **kwargs,
    ):
        Component.__init__(self, css=css, **kwargs)
        self.components = components
        self.type = type
        self.headers = [c.label for c in components]
        self.samples = samples

    def get_template_context(self):
        return {
            "components": [component.get_block_name() for component in self.components],
            "headers": self.headers,
            "samples": self.samples,
            "type": self.type,
            **Component.get_template_context(self),
        }

    def preprocess(self, x: Any) -> Any:
        """
        Any preprocessing needed to be performed on function input.
        """
        if self.type == "index":
            return x
        elif self.type == "values":
            return self.samples[x]


class Interpretation(Component):
    """
    Used to create an interpretation widget for a component.
    """

    def __init__(
        self,
        component: Component,
        *,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        Component.__init__(self, css=css, **kwargs)
        self.component = component

    def get_template_context(self):
        return {
            "component": self.component.get_block_name(),
            "component_props": self.component.get_template_context(),
        }


class StatusTracker(Component):
    """
    Used to indicate status of a function call. Event listeners can bind to a StatusTracker with 'status=' keyword argument.
    """

    def __init__(
        self,
        *,
        cover_container: bool = False,
        css: Optional[CSS] = None,
        **kwargs,
    ):
        """
        Parameters:
        cover_container (bool): If True, will expand to cover parent container while function pending.
        css (dict): optional css parameters for the component
        """
        Component.__init__(self, css=css, **kwargs)
        self.cover_container = cover_container

    def get_template_context(self):
        return {
            "cover_container": self.cover_container,
            **Component.get_template_context(self),
        }


def component(cls_name: str):
    """
    Returns a component or template with the given class name, or raises a ValueError if not found.
    @param cls_name: lower-case string class name of a component
    @return cls: the component class
    """
    import gradio.templates

    components = [
        (name, cls)
        for name, cls in sys.modules[__name__].__dict__.items()
        if isinstance(cls, type)
    ]
    templates = [
        (name, cls)
        for name, cls in gradio.templates.__dict__.items()
        if isinstance(cls, type)
    ]
    for name, cls in components + templates:
        if name.lower() == cls_name.replace("_", "") and issubclass(cls, Component):
            return cls
    raise ValueError(f"No such Component: {cls_name}")


def get_component_instance(comp: str | dict | Component):
    if isinstance(comp, str):
        component_cls = component(comp)
        return component_cls()
    elif isinstance(comp, dict):
        name = comp.pop("name")
        component_cls = component(name)
        return component_cls(**comp)
    elif isinstance(comp, Component):
        return comp
    else:
        raise ValueError(
            f"Component must provided as a `str` or `dict` or `Component` but is {comp}"
        )



import pkg_resources

import gradio.components as components
import gradio.inputs as inputs
import gradio.outputs as outputs
import gradio.processing_utils
from gradio.blocks import Blocks, Column, Row, TabItem, Tabs
from gradio.components import (
    HTML,
    JSON,
    Audio,
    Button,
    Carousel,
    Chatbot,
    Checkbox,
    CheckboxGroup,
    Dataframe,
    Dropdown,
    File,
    Gallery,
    HighlightedText,
    Image,
    KeyValues,
    Label,
    Markdown,
    Model3D,
    Number,
    Plot,
    Radio,
    Slider,
    StatusTracker,
    Textbox,
    Timeseries,
    Variable,
    Video,
    component,
)
from gradio.flagging import (
    CSVLogger,
    FlaggingCallback,
    HuggingFaceDatasetSaver,
    SimpleCSVLogger,
)
from gradio.interface import Interface, TabbedInterface, close_all
from gradio.mix import Parallel, Series
from gradio.css import CSS

current_pkg_version = pkg_resources.require("gradio")[0].version
__version__ = current_pkg_version



from dataclasses import dataclass
from typing import Optional
import re

@dataclass
class CSS:
    alignContent: Optional[str] = None
    alignItems: Optional[str] = None
    alignSelf: Optional[str] = None
    alignmentBaseline: Optional[str] = None
    all: Optional[str] = None
    animation: Optional[str] = None
    animationDelay: Optional[str] = None
    animationDirection: Optional[str] = None
    animationDuration: Optional[str] = None
    animationFillMode: Optional[str] = None
    animationIterationCount: Optional[str] = None
    animationName: Optional[str] = None
    animationPlayState: Optional[str] = None
    animationTimingFunction: Optional[str] = None
    appearance: Optional[str] = None
    ascentOverride: Optional[str] = None
    aspectRatio: Optional[str] = None
    backdropFilter: Optional[str] = None
    backfaceVisibility: Optional[str] = None
    background: Optional[str] = None
    backgroundAttachment: Optional[str] = None
    backgroundBlendMode: Optional[str] = None
    backgroundClip: Optional[str] = None
    backgroundColor: Optional[str] = None
    backgroundImage: Optional[str] = None
    backgroundOrigin: Optional[str] = None
    backgroundPosition: Optional[str] = None
    backgroundPositionX: Optional[str] = None
    backgroundPositionY: Optional[str] = None
    backgroundRepeat: Optional[str] = None
    backgroundRepeatX: Optional[str] = None
    backgroundRepeatY: Optional[str] = None
    backgroundSize: Optional[str] = None
    baselineShift: Optional[str] = None
    blockSize: Optional[str] = None
    border: Optional[str] = None
    borderBlock: Optional[str] = None
    borderBlockColor: Optional[str] = None
    borderBlockEnd: Optional[str] = None
    borderBlockEndColor: Optional[str] = None
    borderBlockEndStyle: Optional[str] = None
    borderBlockEndWidth: Optional[str] = None
    borderBlockStart: Optional[str] = None
    borderBlockStartColor: Optional[str] = None
    borderBlockStartStyle: Optional[str] = None
    borderBlockStartWidth: Optional[str] = None
    borderBlockStyle: Optional[str] = None
    borderBlockWidth: Optional[str] = None
    borderBottom: Optional[str] = None
    borderBottomColor: Optional[str] = None
    borderBottomLeftRadius: Optional[str] = None
    borderBottomRightRadius: Optional[str] = None
    borderBottomStyle: Optional[str] = None
    borderBottomWidth: Optional[str] = None
    borderCollapse: Optional[str] = None
    borderColor: Optional[str] = None
    borderEndEndRadius: Optional[str] = None
    borderEndStartRadius: Optional[str] = None
    borderImage: Optional[str] = None
    borderImageOutset: Optional[str] = None
    borderImageRepeat: Optional[str] = None
    borderImageSlice: Optional[str] = None
    borderImageSource: Optional[str] = None
    borderImageWidth: Optional[str] = None
    borderInline: Optional[str] = None
    borderInlineColor: Optional[str] = None
    borderInlineEnd: Optional[str] = None
    borderInlineEndColor: Optional[str] = None
    borderInlineEndStyle: Optional[str] = None
    borderInlineEndWidth: Optional[str] = None
    borderInlineStart: Optional[str] = None
    borderInlineStartColor: Optional[str] = None
    borderInlineStartStyle: Optional[str] = None
    borderInlineStartWidth: Optional[str] = None
    borderInlineStyle: Optional[str] = None
    borderInlineWidth: Optional[str] = None
    borderLeft: Optional[str] = None
    borderLeftColor: Optional[str] = None
    borderLeftStyle: Optional[str] = None
    borderLeftWidth: Optional[str] = None
    borderRadius: Optional[str] = None
    borderRight: Optional[str] = None
    borderRightColor: Optional[str] = None
    borderRightStyle: Optional[str] = None
    borderRightWidth: Optional[str] = None
    borderSpacing: Optional[str] = None
    borderStartEndRadius: Optional[str] = None
    borderStartStartRadius: Optional[str] = None
    borderStyle: Optional[str] = None
    borderTop: Optional[str] = None
    borderTopColor: Optional[str] = None
    borderTopLeftRadius: Optional[str] = None
    borderTopRightRadius: Optional[str] = None
    borderTopStyle: Optional[str] = None
    borderTopWidth: Optional[str] = None
    borderWidth: Optional[str] = None
    bottom: Optional[str] = None
    boxShadow: Optional[str] = None
    boxSizing: Optional[str] = None
    breakAfter: Optional[str] = None
    breakBefore: Optional[str] = None
    breakInside: Optional[str] = None
    bufferedRendering: Optional[str] = None
    captionSide: Optional[str] = None
    caretColor: Optional[str] = None
    clear: Optional[str] = None
    clip: Optional[str] = None
    clipPath: Optional[str] = None
    clipRule: Optional[str] = None
    color: Optional[str] = None
    colorInterpolation: Optional[str] = None
    colorInterpolationFilters: Optional[str] = None
    colorRendering: Optional[str] = None
    colorScheme: Optional[str] = None
    columnCount: Optional[str] = None
    columnFill: Optional[str] = None
    columnGap: Optional[str] = None
    columnRule: Optional[str] = None
    columnRuleColor: Optional[str] = None
    columnRuleStyle: Optional[str] = None
    columnRuleWidth: Optional[str] = None
    columnSpan: Optional[str] = None
    columnWidth: Optional[str] = None
    columns: Optional[str] = None
    contain: Optional[str] = None
    containIntrinsicSize: Optional[str] = None
    content: Optional[str] = None
    contentVisibility: Optional[str] = None
    counterIncrement: Optional[str] = None
    counterReset: Optional[str] = None
    counterSet: Optional[str] = None
    cursor: Optional[str] = None
    cx: Optional[str] = None
    cy: Optional[str] = None
    d: Optional[str] = None
    descentOverride: Optional[str] = None
    direction: Optional[str] = None
    display: Optional[str] = None
    dominantBaseline: Optional[str] = None
    emptyCells: Optional[str] = None
    fallback: Optional[str] = None
    fill: Optional[str] = None
    fillOpacity: Optional[str] = None
    fillRule: Optional[str] = None
    filter: Optional[str] = None
    flex: Optional[str] = None
    flexBasis: Optional[str] = None
    flexDirection: Optional[str] = None
    flexFlow: Optional[str] = None
    flexGrow: Optional[str] = None
    flexShrink: Optional[str] = None
    flexWrap: Optional[str] = None
    float: Optional[str] = None
    floodColor: Optional[str] = None
    floodOpacity: Optional[str] = None
    font: Optional[str] = None
    fontDisplay: Optional[str] = None
    fontFamily: Optional[str] = None
    fontFeatureSettings: Optional[str] = None
    fontKerning: Optional[str] = None
    fontOpticalSizing: Optional[str] = None
    fontSize: Optional[str] = None
    fontStretch: Optional[str] = None
    fontStyle: Optional[str] = None
    fontVariant: Optional[str] = None
    fontVariantCaps: Optional[str] = None
    fontVariantEastAsian: Optional[str] = None
    fontVariantLigatures: Optional[str] = None
    fontVariantNumeric: Optional[str] = None
    fontVariationSettings: Optional[str] = None
    fontWeight: Optional[str] = None
    forcedColorAdjust: Optional[str] = None
    gap: Optional[str] = None
    grid: Optional[str] = None
    gridArea: Optional[str] = None
    gridAutoColumns: Optional[str] = None
    gridAutoFlow: Optional[str] = None
    gridAutoRows: Optional[str] = None
    gridColumn: Optional[str] = None
    gridColumnEnd: Optional[str] = None
    gridColumnGap: Optional[str] = None
    gridColumnStart: Optional[str] = None
    gridGap: Optional[str] = None
    gridRow: Optional[str] = None
    gridRowEnd: Optional[str] = None
    gridRowGap: Optional[str] = None
    gridRowStart: Optional[str] = None
    gridTemplate: Optional[str] = None
    gridTemplateAreas: Optional[str] = None
    gridTemplateColumns: Optional[str] = None
    gridTemplateRows: Optional[str] = None
    height: Optional[str] = None
    hyphens: Optional[str] = None
    imageOrientation: Optional[str] = None
    imageRendering: Optional[str] = None
    inherits: Optional[str] = None
    initialValue: Optional[str] = None
    inlineSize: Optional[str] = None
    inset: Optional[str] = None
    insetBlock: Optional[str] = None
    insetBlockEnd: Optional[str] = None
    insetBlockStart: Optional[str] = None
    insetInline: Optional[str] = None
    insetInlineEnd: Optional[str] = None
    insetInlineStart: Optional[str] = None
    isolation: Optional[str] = None
    justifyContent: Optional[str] = None
    justifyItems: Optional[str] = None
    justifySelf: Optional[str] = None
    left: Optional[str] = None
    letterSpacing: Optional[str] = None
    lightingColor: Optional[str] = None
    lineBreak: Optional[str] = None
    lineGapOverride: Optional[str] = None
    lineHeight: Optional[str] = None
    listStyle: Optional[str] = None
    listStyleImage: Optional[str] = None
    listStylePosition: Optional[str] = None
    listStyleType: Optional[str] = None
    margin: Optional[str] = None
    marginBlock: Optional[str] = None
    marginBlockEnd: Optional[str] = None
    marginBlockStart: Optional[str] = None
    marginBottom: Optional[str] = None
    marginInline: Optional[str] = None
    marginInlineEnd: Optional[str] = None
    marginInlineStart: Optional[str] = None
    marginLeft: Optional[str] = None
    marginRight: Optional[str] = None
    marginTop: Optional[str] = None
    marker: Optional[str] = None
    markerEnd: Optional[str] = None
    markerMid: Optional[str] = None
    markerStart: Optional[str] = None
    mask: Optional[str] = None
    maskType: Optional[str] = None
    maxBlockSize: Optional[str] = None
    maxHeight: Optional[str] = None
    maxInlineSize: Optional[str] = None
    maxWidth: Optional[str] = None
    maxZoom: Optional[str] = None
    minBlockSize: Optional[str] = None
    minHeight: Optional[str] = None
    minInlineSize: Optional[str] = None
    minWidth: Optional[str] = None
    minZoom: Optional[str] = None
    mixBlendMode: Optional[str] = None
    negative: Optional[str] = None
    objectFit: Optional[str] = None
    objectPosition: Optional[str] = None
    offset: Optional[str] = None
    offsetDistance: Optional[str] = None
    offsetPath: Optional[str] = None
    offsetRotate: Optional[str] = None
    opacity: Optional[str] = None
    order: Optional[str] = None
    orientation: Optional[str] = None
    orphans: Optional[str] = None
    outline: Optional[str] = None
    outlineColor: Optional[str] = None
    outlineOffset: Optional[str] = None
    outlineStyle: Optional[str] = None
    outlineWidth: Optional[str] = None
    overflow: Optional[str] = None
    overflowAnchor: Optional[str] = None
    overflowClipMargin: Optional[str] = None
    overflowWrap: Optional[str] = None
    overflowX: Optional[str] = None
    overflowY: Optional[str] = None
    overscrollBehavior: Optional[str] = None
    overscrollBehaviorBlock: Optional[str] = None
    overscrollBehaviorInline: Optional[str] = None
    overscrollBehaviorX: Optional[str] = None
    overscrollBehaviorY: Optional[str] = None
    pad: Optional[str] = None
    padding: Optional[str] = None
    paddingBlock: Optional[str] = None
    paddingBlockEnd: Optional[str] = None
    paddingBlockStart: Optional[str] = None
    paddingBottom: Optional[str] = None
    paddingInline: Optional[str] = None
    paddingInlineEnd: Optional[str] = None
    paddingInlineStart: Optional[str] = None
    paddingLeft: Optional[str] = None
    paddingRight: Optional[str] = None
    paddingTop: Optional[str] = None
    page: Optional[str] = None
    pageBreakAfter: Optional[str] = None
    pageBreakBefore: Optional[str] = None
    pageBreakInside: Optional[str] = None
    pageOrientation: Optional[str] = None
    paintOrder: Optional[str] = None
    perspective: Optional[str] = None
    perspectiveOrigin: Optional[str] = None
    placeContent: Optional[str] = None
    placeItems: Optional[str] = None
    placeSelf: Optional[str] = None
    pointerEvents: Optional[str] = None
    position: Optional[str] = None
    prefix: Optional[str] = None
    quotes: Optional[str] = None
    r: Optional[str] = None
    range: Optional[str] = None
    resize: Optional[str] = None
    right: Optional[str] = None
    rowGap: Optional[str] = None
    rubyPosition: Optional[str] = None
    rx: Optional[str] = None
    ry: Optional[str] = None
    scrollBehavior: Optional[str] = None
    scrollMargin: Optional[str] = None
    scrollMarginBlock: Optional[str] = None
    scrollMarginBlockEnd: Optional[str] = None
    scrollMarginBlockStart: Optional[str] = None
    scrollMarginBottom: Optional[str] = None
    scrollMarginInline: Optional[str] = None
    scrollMarginInlineEnd: Optional[str] = None
    scrollMarginInlineStart: Optional[str] = None
    scrollMarginLeft: Optional[str] = None
    scrollMarginRight: Optional[str] = None
    scrollMarginTop: Optional[str] = None
    scrollPadding: Optional[str] = None
    scrollPaddingBlock: Optional[str] = None
    scrollPaddingBlockEnd: Optional[str] = None
    scrollPaddingBlockStart: Optional[str] = None
    scrollPaddingBottom: Optional[str] = None
    scrollPaddingInline: Optional[str] = None
    scrollPaddingInlineEnd: Optional[str] = None
    scrollPaddingInlineStart: Optional[str] = None
    scrollPaddingLeft: Optional[str] = None
    scrollPaddingRight: Optional[str] = None
    scrollPaddingTop: Optional[str] = None
    scrollSnapAlign: Optional[str] = None
    scrollSnapStop: Optional[str] = None
    scrollSnapType: Optional[str] = None
    shapeImageThreshold: Optional[str] = None
    shapeMargin: Optional[str] = None
    shapeOutside: Optional[str] = None
    shapeRendering: Optional[str] = None
    size: Optional[str] = None
    sizeAdjust: Optional[str] = None
    speak: Optional[str] = None
    speakAs: Optional[str] = None
    src: Optional[str] = None
    stopColor: Optional[str] = None
    stopOpacity: Optional[str] = None
    stroke: Optional[str] = None
    strokeDasharray: Optional[str] = None
    strokeDashoffset: Optional[str] = None
    strokeLinecap: Optional[str] = None
    strokeLinejoin: Optional[str] = None
    strokeMiterlimit: Optional[str] = None
    strokeOpacity: Optional[str] = None
    strokeWidth: Optional[str] = None
    suffix: Optional[str] = None
    symbols: Optional[str] = None
    syntax: Optional[str] = None
    system: Optional[str] = None
    tabSize: Optional[str] = None
    tableLayout: Optional[str] = None
    textAlign: Optional[str] = None
    textAlignLast: Optional[str] = None
    textAnchor: Optional[str] = None
    textCombineUpright: Optional[str] = None
    textDecoration: Optional[str] = None
    textDecorationColor: Optional[str] = None
    textDecorationLine: Optional[str] = None
    textDecorationSkipInk: Optional[str] = None
    textDecorationStyle: Optional[str] = None
    textDecorationThickness: Optional[str] = None
    textIndent: Optional[str] = None
    textOrientation: Optional[str] = None
    textOverflow: Optional[str] = None
    textRendering: Optional[str] = None
    textShadow: Optional[str] = None
    textSizeAdjust: Optional[str] = None
    textTransform: Optional[str] = None
    textUnderlineOffset: Optional[str] = None
    textUnderlinePosition: Optional[str] = None
    top: Optional[str] = None
    touchAction: Optional[str] = None
    transform: Optional[str] = None
    transformBox: Optional[str] = None
    transformOrigin: Optional[str] = None
    transformStyle: Optional[str] = None
    transition: Optional[str] = None
    transitionDelay: Optional[str] = None
    transitionDuration: Optional[str] = None
    transitionProperty: Optional[str] = None
    transitionTimingFunction: Optional[str] = None
    unicodeBidi: Optional[str] = None
    unicodeRange: Optional[str] = None
    userSelect: Optional[str] = None
    userZoom: Optional[str] = None
    vectorEffect: Optional[str] = None
    verticalAlign: Optional[str] = None
    visibility: Optional[str] = None
    webkitAlignContent: Optional[str] = None
    webkitAlignItems: Optional[str] = None
    webkitAlignSelf: Optional[str] = None
    webkitAnimation: Optional[str] = None
    webkitAnimationDelay: Optional[str] = None
    webkitAnimationDirection: Optional[str] = None
    webkitAnimationDuration: Optional[str] = None
    webkitAnimationFillMode: Optional[str] = None
    webkitAnimationIterationCount: Optional[str] = None
    webkitAnimationName: Optional[str] = None
    webkitAnimationPlayState: Optional[str] = None
    webkitAnimationTimingFunction: Optional[str] = None
    webkitAppRegion: Optional[str] = None
    webkitAppearance: Optional[str] = None
    webkitBackfaceVisibility: Optional[str] = None
    webkitBackgroundClip: Optional[str] = None
    webkitBackgroundOrigin: Optional[str] = None
    webkitBackgroundSize: Optional[str] = None
    webkitBorderAfter: Optional[str] = None
    webkitBorderAfterColor: Optional[str] = None
    webkitBorderAfterStyle: Optional[str] = None
    webkitBorderAfterWidth: Optional[str] = None
    webkitBorderBefore: Optional[str] = None
    webkitBorderBeforeColor: Optional[str] = None
    webkitBorderBeforeStyle: Optional[str] = None
    webkitBorderBeforeWidth: Optional[str] = None
    webkitBorderBottomLeftRadius: Optional[str] = None
    webkitBorderBottomRightRadius: Optional[str] = None
    webkitBorderEnd: Optional[str] = None
    webkitBorderEndColor: Optional[str] = None
    webkitBorderEndStyle: Optional[str] = None
    webkitBorderEndWidth: Optional[str] = None
    webkitBorderHorizontalSpacing: Optional[str] = None
    webkitBorderImage: Optional[str] = None
    webkitBorderRadius: Optional[str] = None
    webkitBorderStart: Optional[str] = None
    webkitBorderStartColor: Optional[str] = None
    webkitBorderStartStyle: Optional[str] = None
    webkitBorderStartWidth: Optional[str] = None
    webkitBorderTopLeftRadius: Optional[str] = None
    webkitBorderTopRightRadius: Optional[str] = None
    webkitBorderVerticalSpacing: Optional[str] = None
    webkitBoxAlign: Optional[str] = None
    webkitBoxDecorationBreak: Optional[str] = None
    webkitBoxDirection: Optional[str] = None
    webkitBoxFlex: Optional[str] = None
    webkitBoxOrdinalGroup: Optional[str] = None
    webkitBoxOrient: Optional[str] = None
    webkitBoxPack: Optional[str] = None
    webkitBoxReflect: Optional[str] = None
    webkitBoxShadow: Optional[str] = None
    webkitBoxSizing: Optional[str] = None
    webkitClipPath: Optional[str] = None
    webkitColumnBreakAfter: Optional[str] = None
    webkitColumnBreakBefore: Optional[str] = None
    webkitColumnBreakInside: Optional[str] = None
    webkitColumnCount: Optional[str] = None
    webkitColumnGap: Optional[str] = None
    webkitColumnRule: Optional[str] = None
    webkitColumnRuleColor: Optional[str] = None
    webkitColumnRuleStyle: Optional[str] = None
    webkitColumnRuleWidth: Optional[str] = None
    webkitColumnSpan: Optional[str] = None
    webkitColumnWidth: Optional[str] = None
    webkitColumns: Optional[str] = None
    webkitFilter: Optional[str] = None
    webkitFlex: Optional[str] = None
    webkitFlexBasis: Optional[str] = None
    webkitFlexDirection: Optional[str] = None
    webkitFlexFlow: Optional[str] = None
    webkitFlexGrow: Optional[str] = None
    webkitFlexShrink: Optional[str] = None
    webkitFlexWrap: Optional[str] = None
    webkitFontFeatureSettings: Optional[str] = None
    webkitFontSmoothing: Optional[str] = None
    webkitHighlight: Optional[str] = None
    webkitHyphenateCharacter: Optional[str] = None
    webkitJustifyContent: Optional[str] = None
    webkitLineBreak: Optional[str] = None
    webkitLineClamp: Optional[str] = None
    webkitLocale: Optional[str] = None
    webkitLogicalHeight: Optional[str] = None
    webkitLogicalWidth: Optional[str] = None
    webkitMarginAfter: Optional[str] = None
    webkitMarginBefore: Optional[str] = None
    webkitMarginEnd: Optional[str] = None
    webkitMarginStart: Optional[str] = None
    webkitMask: Optional[str] = None
    webkitMaskBoxImage: Optional[str] = None
    webkitMaskBoxImageOutset: Optional[str] = None
    webkitMaskBoxImageRepeat: Optional[str] = None
    webkitMaskBoxImageSlice: Optional[str] = None
    webkitMaskBoxImageSource: Optional[str] = None
    webkitMaskBoxImageWidth: Optional[str] = None
    webkitMaskClip: Optional[str] = None
    webkitMaskComposite: Optional[str] = None
    webkitMaskImage: Optional[str] = None
    webkitMaskOrigin: Optional[str] = None
    webkitMaskPosition: Optional[str] = None
    webkitMaskPositionX: Optional[str] = None
    webkitMaskPositionY: Optional[str] = None
    webkitMaskRepeat: Optional[str] = None
    webkitMaskRepeatX: Optional[str] = None
    webkitMaskRepeatY: Optional[str] = None
    webkitMaskSize: Optional[str] = None
    webkitMaxLogicalHeight: Optional[str] = None
    webkitMaxLogicalWidth: Optional[str] = None
    webkitMinLogicalHeight: Optional[str] = None
    webkitMinLogicalWidth: Optional[str] = None
    webkitOpacity: Optional[str] = None
    webkitOrder: Optional[str] = None
    webkitPaddingAfter: Optional[str] = None
    webkitPaddingBefore: Optional[str] = None
    webkitPaddingEnd: Optional[str] = None
    webkitPaddingStart: Optional[str] = None
    webkitPerspective: Optional[str] = None
    webkitPerspectiveOrigin: Optional[str] = None
    webkitPerspectiveOriginX: Optional[str] = None
    webkitPerspectiveOriginY: Optional[str] = None
    webkitPrintColorAdjust: Optional[str] = None
    webkitRtlOrdering: Optional[str] = None
    webkitRubyPosition: Optional[str] = None
    webkitShapeImageThreshold: Optional[str] = None
    webkitShapeMargin: Optional[str] = None
    webkitShapeOutside: Optional[str] = None
    webkitTapHighlightColor: Optional[str] = None
    webkitTextCombine: Optional[str] = None
    webkitTextDecorationsInEffect: Optional[str] = None
    webkitTextEmphasis: Optional[str] = None
    webkitTextEmphasisColor: Optional[str] = None
    webkitTextEmphasisPosition: Optional[str] = None
    webkitTextEmphasisStyle: Optional[str] = None
    webkitTextFillColor: Optional[str] = None
    webkitTextOrientation: Optional[str] = None
    webkitTextSecurity: Optional[str] = None
    webkitTextSizeAdjust: Optional[str] = None
    webkitTextStroke: Optional[str] = None
    webkitTextStrokeColor: Optional[str] = None
    webkitTextStrokeWidth: Optional[str] = None
    webkitTransform: Optional[str] = None
    webkitTransformOrigin: Optional[str] = None
    webkitTransformOriginX: Optional[str] = None
    webkitTransformOriginY: Optional[str] = None
    webkitTransformOriginZ: Optional[str] = None
    webkitTransformStyle: Optional[str] = None
    webkitTransition: Optional[str] = None
    webkitTransitionDelay: Optional[str] = None
    webkitTransitionDuration: Optional[str] = None
    webkitTransitionProperty: Optional[str] = None
    webkitTransitionTimingFunction: Optional[str] = None
    webkitUserDrag: Optional[str] = None
    webkitUserModify: Optional[str] = None
    webkitUserSelect: Optional[str] = None
    webkitWritingMode: Optional[str] = None
    whiteSpace: Optional[str] = None
    widows: Optional[str] = None
    width: Optional[str] = None
    willChange: Optional[str] = None
    wordBreak: Optional[str] = None
    wordSpacing: Optional[str] = None
    wordWrap: Optional[str] = None
    writingMode: Optional[str] = None
    x: Optional[str] = None
    y: Optional[str] = None
    zIndex: Optional[str] = None
    zoom: Optional[str] = None

    def render(self):
        output = {}
        for attr, value in self.__dict__.items():
            if type(value) == str:
                kebab_case = re.sub(r'([a-z0-9]|(?=[A-Z]))([A-Z])', r'\1-\2', attr).lower()
                output[kebab_case] = value
            elif type(value) == CSS:
                output[attr] = value.render()
        return output

@dataclass
class CSSWithLabel(CSS):
    labelStyle: Optional[CSS] = None
